# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .plugins import PluginHandler, Plugin
from .feed import allfeeds
from .hooks import call_hook

from multiprocessing import cpu_count
from threading import Thread

import feedparser
import traceback
import urllib.parse
import urllib.request
import urllib.error
import logging
import socket
import json
import time

log = logging.getLogger("CANTO-FETCH")

# Function to pass to json.dumps to strip non-serializable data
def json_ignore(x):
    return None

class DaemonFetchThreadPlugin(Plugin):
    pass

# This is the first time I've ever had a need for multiple inheritance.
# I'm not sure if that's a good thing or not =)

class CantoFetchThread(PluginHandler, Thread):
    def __init__(self, feed, fromdisk):
        PluginHandler.__init__(self)
        Thread.__init__(self, name="Fetch: %s" % feed.URL)
        self.daemon = True

        self.plugin_class = DaemonFetchThreadPlugin
        self.update_plugin_lookups()

        # feedparser honors this value, want to avoid hung feeds when the
        # internet connection is flaky

        socket.setdefaulttimeout(30)

        self.feed = feed
        self.fromdisk = fromdisk

    def run(self):

        # Initial load, just feed.index grab from disk.

        if self.fromdisk:
            self.feed.index({"entries" : []})
            return

        self.feed.last_update = time.time()

        # Otherwise, actually try to get an update.

        extra_headers = { 'User-Agent' :\
                'Canto/0.9.0 + http://codezen.org/canto-ng'}

        try:
            result = None
            # Passworded Feed
            if self.feed.username or self.feed.password:
                domain = urllib.parse.urlparse(self.feed.URL)[1]
                man = urllib.request.HTTPPasswordMgrWithDefaultRealm()
                auth = urllib.request.HTTPBasicAuthHandler(man)
                auth.handler_order = 490
                auth.add_password(None, domain, self.feed.username,
                        self.feed.password)

                try:
                    result = feedparser.parse(self.feed.URL, handlers=[auth],
                            request_headers = extra_headers)
                except:
                    # And, failing that, Digest Authentication
                    man = urllib.request.HTTPPasswordMgrWithDefaultRealm()
                    auth = urllib.request.HTTPDigestAuthHandler(man)
                    auth.handler_order = 490
                    auth.add_password(None, domain, self.feed.username,
                            self.feed.password)
                    result = feedparser.parse(self.feed.URL, handlers=[auth],
                            request_headers = extra_headers)

            # No password
            else:
                result = feedparser.parse(self.feed.URL,
                        request_headers = extra_headers)

            update_contents = result
        except Exception as e:
            log.error("ERROR: try to parse %s, got %s" % (self.feed.URL, e))
            return

        # Allow DaemonFetchThreadPlugins to do any sort of fetch stuff Doing
        # this before any other processing allows us to have plugins that
        # totally override the standard fetch.

        for attr in list(self.plugin_attrs.keys()):
            if not attr.startswith("fetch_"):
                continue

            try:
                a = getattr(self, attr)
                a(feed = self.feed, newcontent = update_contents)
            except:
                log.error("Error running fetch thread plugin")
                log.error(traceback.format_exc())

        log.debug("Plugins complete.")

        # Interpret feedparser's bozo_exception, if there was an
        # error that resulted in no content, it's the same as
        # any other broken feed.

        if "bozo_exception" in update_contents:
            if update_contents["bozo_exception"] == urllib.error.URLError:
                log.error("ERROR: couldn't grab %s : %s" %\
                        (self.feed.URL,\
                        update_contents["bozo_exception"].reason))
                return
            elif len(update_contents["entries"]) == 0:
                log.error("No content in %s: %s" %\
                        (self.feed.URL,\
                        update_contents["bozo_exception"]))
                return

            # Replace it if we ignore it, since exceptions
            # are not pickle-able.

            update_contents["bozo_exception"] = None

        # Update timestamp
        update_contents["canto_update"] = self.feed.last_update

        update_contents = json.loads(json.dumps(update_contents, default=json_ignore))

        log.debug("Parsed %s", self.feed.URL)

        # This handles it's own locking
        self.feed.index(update_contents)

class CantoFetch():
    def __init__(self, shelf):
        self.shelf = shelf
        self.deferred = []
        self.threads = []
        self.thread_limit = cpu_count()
        log.debug("Thread Limit: %s", self.thread_limit)

    def needs_update(self, feed):
        passed = time.time() - feed.last_update
        if passed < feed.rate * 60:
            return False
        return True

    def still_working(self, URL):
        for thread, workingURL in self.threads:
            if URL == workingURL:
                return True
        return False

    def _start_one(self, feed, fromdisk):
        if len(self.threads) >= self.thread_limit:
            return False

        # If feed is stopped/dead, pretend like we did the work but don't
        # resurrect tags

        if feed.stopped:
            return True

        thread = CantoFetchThread(feed, fromdisk)
        thread.start()
        log.debug("Started thread for feed %s", feed)
        self.threads.append((thread, feed.URL))
        return True

    def fetch(self, force, fromdisk):
        for feed, fd in self.deferred[:]:
            if self._start_one(feed, fd):
                log.debug("No longer deferred")
                self.deferred = self.deferred[1:]
            else:
                return

        for feed in allfeeds.get_feeds():
            if not force and not self.needs_update(feed):
                continue

            if self.still_working(feed.URL):
                continue

            if not self._start_one(feed, fromdisk):
                log.debug("Deferring %s %s", feed, fromdisk)
                self.deferred.append((feed, fromdisk))

    def reap(self, force=False):
        work_done = False
        newthreads = []

        for thread, URL in self.threads:
            if not force and thread.is_alive():
                newthreads.append((thread, URL))
                continue
            work_done = True
            thread.join()

        self.threads = newthreads

        if work_done and self.threads == []:
            self.shelf.sync()



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from threading import Lock
import logging
import socket
import select
import errno
import getopt
import struct
import shlex
import json
import time
import sys
import os

log = logging.getLogger('SOCKET')

class CantoSocket:
    def __init__(self, socket_name, **kwargs):

        self.socket_name = socket_name

        if "server" in kwargs and kwargs["server"]:
            self.server = True
        else:
            self.server = False

        if "port" in kwargs:
            self.port = kwargs["port"]
        else:
            self.port = 0

        if "interface" in kwargs:
            self.interface = kwargs["interface"]
        else:
            self.interface = ''

        if "address" in kwargs:
            self.address = kwargs["address"]
        else:
            self.address = None

        self.sockets = []
        self.read_locks = {}
        self.write_locks = {}
        self.write_frags = {}

        self.connect()

    # Handle options common to all servers and clients

    def common_args(self, extrashort = "", extralong = [], version = ""):
        self.verbosity = 0
        self.port = -1
        self.addr = None
        self.disabled_plugins = []
        self.enabled_plugins = []
        self.plugin_default = True

        try:
            optlist, sys.argv =\
                getopt.getopt(sys.argv[1:], 'D:p:a:vV' + extrashort, ["dir=",
                "port=", "address=","version", "noplugins","enableplugins=",
                "disableplugins="] + extralong)

        except getopt.GetoptError as e:
            log.error("Error: %s" % e.msg)
            return -1

        old_path = os.path.expanduser("~/.canto-ng")

        if os.path.exists(old_path):
            self.conf_dir = old_path
        else:
            if "XDG_CONFIG_HOME" in os.environ:
                xdg_path = os.environ["XDG_CONFIG_HOME"]
            else:
                xdg_path = "~/.config"

            xdg_path = os.path.expanduser(xdg_path)
            self.conf_dir = xdg_path + "/canto"

        self.location_args = []

        for opt, arg in optlist:
            if opt in [ "-D", "--dir"]:
                self.conf_dir = os.path.expanduser(arg)
                self.conf_dir = os.path.realpath(self.conf_dir)
                self.location_args += [ opt, arg ]

            elif opt in ["-V", "--version"]:
                print(version)
                sys.exit(0)

            elif opt in ["-v"]:
                self.verbosity += 1

            elif opt in [ "-p", "--port"]:
                try:
                    self.port = int(arg)
                    if self.port < 0:
                        raise Exception
                except:
                    log.error("Error: Port must be >0 integer.")
                    return -1

                # Assume loopback if address hasn't been set yet.
                if self.addr == None:
                    self.addr = "127.0.0.1"

                self.location_args += [ opt, arg ]

            elif opt in [ "-a", "--address"]:
                self.addr = arg
                self.location_args += [ opt, arg ]

            elif opt in ['--noplugins']:
                self.plugin_default = False

            elif opt in ['--disableplugins']:
                self.disabled_plugins = shlex.split(arg)

            elif opt in ['--enableplugins']:
                self.enabled_plugins = shlex.split(arg)

        self.socket_path = self.conf_dir + "/.canto_socket"

        return optlist

    # Server setup, potentially both unix and inet sockets.
    def connect(self):
        if self.server:
            if self.socket_name:
                # Remove old unix socket.
                if os.path.exists(self.socket_name):
                    os.remove(self.socket_name)

                # Setup new socket.
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                sock.setblocking(0)
                sock.bind(self.socket_name)
                sock.listen(5)

            # Net socket setup.
            if self.port > 0:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setblocking(0)

                sock.bind((self.interface, self.port))
                sock.listen(5)

        # Client setup, can only do unix or inet, not both.

        else:
            if self.address and self.port > 0:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                addr = (self.address, self.port)
            else:
                sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
                addr = self.socket_name

            tries = 10

            while tries > 0:
                try:
                    sock.connect(addr)
                    break
                except Exception as e:
                    if e.args[0] != errno.ECONNREFUSED or tries == 1:
                        raise
                time.sleep(1)
                tries -= 1

        self.sockets.append(sock)
        self.read_locks[sock] = Lock()
        self.write_locks[sock] = Lock()
        self.write_frags[sock] = None
        return sock

    # Setup poll.poll() object to watch for read status on conn.
    def read_mode(self, poll, conn):
        poll.register(conn.fileno(),\
                select.POLLIN | select.POLLHUP | select.POLLERR |\
                select.POLLPRI)

    # Setup poll.poll() object to watch for write status on conn.
    def write_mode(self, poll, conn):
        poll.register(conn.fileno(),\
                select.POLLOUT | select.POLLHUP | select.POLLERR |\
                select.POLLNVAL)

    # Take raw data, return (cmd, args) tuple or None if not enough data.
    def parse(self, conn, data):
        try:
            cmd, args = eval(repr(json.loads(data)), {}, {})
        except:
            log.error("Failed to parse message: %s" % data)
        else:
            log.debug("\n\nRead:\n%s", json.dumps((cmd, args), indent=4, sort_keys=True))
            return (cmd, args)

    def do_read(self, conn, timeout=None):
        while True:
            to = timeout
            if self.write_frags[conn] != None:
                if to == None:
                    to = 500
                self.do_write(conn, None, None)

            self.read_locks[conn].acquire()
            r = self._do_read(conn, to)
            self.read_locks[conn].release()

            if r == select.POLLHUP:
                self.disconnected(conn)
            elif r == None and timeout == None:
                continue
            return r

    def _do_read(self, conn, timeout):
        poll = select.poll()

        try:
            self.read_mode(poll, conn)
        except:
            log.error("Error putting conn in read mode.")
            log.error("Interpreting as HUP")
            return select.POLLHUP

        # We only care about the first (only) descriptor's event
        try:
            p = poll.poll(timeout)
        except select.error as e:
            if e.args[0] == errno.EINTR:
                return
            log.debug("Raising error: %s", e[1])
            raise

        if timeout and not p:
            return

        e = p[0][1]

        log.debug("E: %d", e)
        if e & select.POLLERR:
            log.debug("Read ERR")
            return select.POLLHUP
        if e & (select.POLLIN | select.POLLPRI):
            message = b""

            try:
                size_bytes = conn.recv(8)
                if not size_bytes:
                    log.debug("No bytes - HUP")
                    return select.POLLHUP
            except:
                log.debug("Couldn't get size, interpreting as HUP\n")
                return select.POLLHUP

            size = struct.unpack('!q', size_bytes)[0]

            while size:
                try:
                    frag = conn.recv(min((4096, size)))
                    size -= len(frag)
                    message += frag
                except Exception as e:
                    if e.args[0] == errno.EINTR:
                        continue

                    log.error("Error receiving: %s" % e)
                    log.error("Interpreting as HUP")
                    return select.POLLHUP

            # Never get POLLRDHUP on INET sockets, so
            # use POLLIN with no data as POLLHUP

            if not message:
                log.debug("Read POLLIN with no data")
                return select.POLLHUP

            return self.parse(conn, message.decode())

        # Parse POLLHUP last so if we still got POLLIN, any data
        # is still retrieved from the socket.
        if e & select.POLLHUP:
            log.debug("Read HUP")
            return select.POLLHUP

        # Non-empty, but not anything we're interested in?
        log.debug("Unknown poll.poll() return")
        return select.POLLHUP

    # Writes a (cmd, args) to a single connection, returns:
    # 1) None if the write completed.
    # 2) select.POLLHUP is the connection is dead.

    def do_write(self, conn, cmd, args):

        # conn could be missing when the connection monitor thread has already
        # cleaned up a connection (i.e. saw it close) before the response
        # thread finishes sending it's response. So, instead of having response
        # threads lock connections (ouch), or deferring the processing of the
        # POLLHUP response, just detect when it's using stale data and ignore
        # it.

        try:
            wlock = self.write_locks[conn]
        except KeyError as e:
            log.debug("conn not in write_locks %s" % e)
            return

        # If we're just flushing data, we shouldn't hang on these:

        if cmd == None:
            if not wlock.acquire(False):
                return
        else:
            wlock.acquire()

        r, frag = self._do_write(conn, cmd, args, self.write_frags[conn])
        wlock.release()

        if r == select.POLLHUP:
            self.disconnected(conn)
        elif r == errno.EINTR:
            self.write_frags[conn] = frag
        else:
            self.write_frags[conn] = None

        return r

    def _do_write(self, conn, cmd, args, frag):
        log.debug("\n\nWrite:\n%s\n", json.dumps((cmd, args), indent=4, sort_keys=True))

        tosend = b""

        if cmd:
            message = json.dumps((cmd, args)).encode("UTF-8")
            size = struct.pack("!q", len(message))
            tosend = size + message

        if frag:
            tosend = frag + tosend

        while tosend:
            poll = select.poll()

            try:
                self.write_mode(poll, conn)
            except:
                log.error("Error putting conn in write mode.")
                log.error("Interpreting as HUP")
                return (select.POLLHUP, 0)

            try:
                p = poll.poll(1)
            except select.error as e:
                if e.args[0] == errno.EINTR:
                    return (errno.EINTR, tosend)
                log.error("Raising error: %s" % e[1])
                raise

            if p == []:
                log.debug("poll timed out")
                return (errno.EINTR, tosend)

            e = p[0][1]

            if e & select.POLLHUP:
                log.debug("Write HUP")
                return (select.POLLHUP, 0)
            if e & select.POLLNVAL:
                log.debug("Write NVAL")
                return (select.POLLHUP, 0)
            if e & select.POLLERR:
                log.debug("Write ERR")
                return (select.POLLHUP, 0)
            if e & select.POLLOUT:
                try:
                    sent = conn.send(tosend)
                except Exception as e:
                    if e.args[0] == errno.EINTR:
                        return (errno.EINTR, tosend)
                    log.error("Error sending: %s" % e[1])
                    log.error("Interpreting as HUP")
                    return (select.POLLHUP, 0)

                tosend = tosend[sent:]
                log.debug("Sent %d bytes.", sent)

        return (None, 0)

    def disconnected(self, conn):
        del self.read_locks[conn]
        del self.write_locks[conn]
        del self.write_frags[conn]



# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .feed import allfeeds
from .tag import alltags

import logging
import re

log = logging.getLogger("TRANSFORM")

transform_locals = { }

# A Transform is generically any form of manipulation of the number of items
# (filter) or order of those items (sort) based on some criteria.

# The CantoTransform class serves as the base of all Transforms. It takes the
# elements returned by a class' `needed_attributes()`, populates a dict of
# these elements from cache/disk, and then gives them to the `transform()` call.

class CantoTransform():
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return self.name

    # This is called with the feeds already read locked.

    def __call__(self, tag):
        a = {}
        f = allfeeds.items_to_feeds(tag)
        needed = self.needed_attributes(tag)

        for feed in f:
            attrs = {}
            for i in f[feed]:
                attrs[i] = needed
            a.update(feed.get_attributes(f[feed], attrs))

        for item in tag[:]:
            if item not in a.keys():
                log.warn("Missing attributes for %s" % item)
                tag.remove(item)

        return self.transform(tag, a)

    def needed_attributes(self, tag):
        return []

    def transform(self, items, attrs):
        return items

# A StateFilter will filter out items that match a particular state. Supports
# using "-tag" to indicate to filter out those missing the tag.

class StateFilter(CantoTransform):
    def __init__(self, state):
        CantoTransform.__init__(self, "Filter state: %s" % state)
        self.state = state

    def needed_attributes(self, tag):
        return ["canto-state"]

    def transform(self, items, attrs):
        if self.state[0] == "-":
            state = self.state[1:]
            keep = True
        else:
            state = self.state
            keep = False

        return [ i for i in items if \
                (state in attrs[i]["canto-state"]) == keep]

# Filter out items whose [attribute] content matches an arbitrary regex.

class ContentFilterRegex(CantoTransform):
    def __init__(self, attribute, regex):
        CantoTransform.__init__(self, "Filter %s in %s" % (attribute, regex))
        self.attribute = attribute
        try:
            self.match = re.compile(regex)
        except:
            self.match = None
            log.error("Couldn't compile regex: %s" % regex)

    def needed_attributes(self, tag):
        if not self.match:
            return []
        return [ self.attribute ]

    def transform(self, items, attrs):
        if not self.match:
            return items

        r = []
        for item in items:
            a = attrs[item]
            if self.attribute not in a:
                r.append(item)
                continue
            if type(a[self.attribute]) != str:
                log.error("Can't match non-string!")
                continue

            if not self.match.match(a[self.attribute]):
                r.append(item)
        return r

# Simple basic-string abstraction of the above.

class ContentFilter(ContentFilterRegex):
    def __init__(self, attribute, string):
        string = ".*" + re.escape(string) + ".*"
        ContentFilterRegex.__init__(self, attribute, string)

class SortTransform(CantoTransform):
    def __init__(self, name, attr):
        CantoTransform.__init__(self, name)
        self.attr = attr

    def needed_attributes(self, tag):
        return [ self.attr ]

    def transform(self, items, attrs):
        r = [ ( attrs[item][self.attr], item ) for item in items ]
        r.sort()
        return [ item[1] for item in r ]

# Meta-filter for AND
class AllTransform(CantoTransform):
    def __init__(self, *args):
        name = "("
        for i, t in enumerate(args):
            if i > 0:
                name += " AND "
            if hasattr(t, "name"):
                name += t.name
            else:
                name += "Unknown"

        name += ")"
        CantoTransform.__init__(self, name)
        self.transforms = args

    def needed_attributes(self, tag):
        needed = []
        for t in self.transforms:
            for a in t.needed_attributes(tag):
                if a not in needed:
                    needed.append(a)
        return needed

    def transform(self, items, attrs):
        good_items = items[:]
        for t in self.transforms:
            good_items = t.transform(good_items, attrs)
            if not good_items:
                break
        return good_items

class AnyTransform(CantoTransform):
    def __init__(self, *args):
        name = "("
        for i, t in enumerate(args):
            if i > 0:
                name += " OR "
            if hasattr(t, "name"):
                name += t.name
            else:
                name += "Unknown"
        name += ")"
        CantoTransform.__init__(self, name)
        self.transforms = args

    def needed_attributes(self, tag):
        needed = []
        for t in self.transforms:
            for a in t.needed_attributes(tag):
                if a not in needed:
                    needed.append(a)
        return needed

    def transform(self, items, attrs):
        good_items = []
        per_transform = []

        for t in self.transforms:
            per_transform.append(t.transform(items, attrs))

        for pt in per_transform:
            for item in pt:
                if item not in good_items:
                    good_items.append(item)
        return good_items

class InTags(CantoTransform):
    def __init__(self, *args):
        name = "in tags: %s" % (args,)

        CantoTransform.__init__(self, name)
        self.tags = args

    def needed_attributes(self, tag):
        return []

    def transform(self, items, attrs):
        good = []

        for item in items:
            for itag in alltags.items_to_tags([item]):
                if itag in self.tags:
                    good.append(item)
                    break

        return good

class ItemLimit(CantoTransform):
    def __init__(self, num):
        if type(num) != int:
            log.error("ItemLimit must be called with a numerical argument")
            self.limit = 0
            return
        else:
            self.limit = num

        self.name="Limit %d items" % self.limit

    def transform(self, items, attrs):
        # Shortcut if failed init
        if self.limit == 0:
            return items

        return items[:self.limit]

# Transform_locals is a list of elements that we pass to the eval() call when
# evaluating a transform line from the config. Passing these into the local
# scope allows simple filters to be created on the fly.

transform_locals["StateFilter"] = StateFilter
transform_locals["ContentFilterRegex"] = ContentFilterRegex
transform_locals["ContentFilter"] = ContentFilter
transform_locals["All"] = AllTransform
transform_locals["Any"] = AnyTransform
transform_locals["InTags"] = InTags
transform_locals["ItemLimit"] = ItemLimit

transform_locals["filter_read"] = StateFilter("read")
transform_locals["sort_alphabetical"] =\
        SortTransform("Sort Alphabetical", "title")

# So now lines line `global_transform = ContentFilter('title', 'AMA')` can be
# simply, safely, parsed with the Python interpreter. As well as supporting the
# simple syntax `global_transform = filter_read` etc.

# This code will throw an exception if it's invalid, so calling code must be
# prepared.

def eval_transform(transform_name):
    return eval(transform_name, {}, transform_locals)



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .feed import wlock_feeds
from .hooks import call_hook

import tempfile
import logging
import shutil
import json
import gzip
import time
import os

log = logging.getLogger("SHELF")

class CantoShelf():
    def __init__(self, filename):
        self.filename = filename

        self.cache = {}

        self.open()

    def check_control_data(self):
        if "control" not in self.cache:
            self.cache["control"] = {}

        for ctrl_field in ["canto-modified","canto-user-modified"]:
            if ctrl_field not in self.cache["control"]:
                self.cache["control"][ctrl_field] = 0

    @wlock_feeds
    def open(self):
        call_hook("daemon_db_open", [self.filename])

        if not os.path.exists(self.filename):
            fp = gzip.open(self.filename, "wt", 9, "UTF-8")
            json.dump(self.cache, fp)
            fp.close()
        else:
            fp = gzip.open(self.filename, "rt", 9, "UTF-8")
            try:
                self.cache = json.load(fp)
            except:
                log.info("Failed to JSON load, old shelf?")
                try:
                    import shelve
                    s = shelve.open(self.filename, "r")
                    for key in s:
                        self.cache[key] = s[key]
                except Exception as e:
                    log.error("Failed to migrate old shelf: %s", e)
                    try:
                        f = open(self.filename)
                        data = f.read()
                        f.close()
                        log.error("BAD DATA: [%s]" % data)
                    except Exception as e:
                        log.error("Couldn't even read data? %s" % e)
                        pass
                    log.error("Carrying on with empty shelf")
                    self.cache = {}
                else:
                    log.info("Migrated old shelf")
            finally:
                fp.close()

        self.check_control_data()

    def __setitem__(self, name, value):
        self.cache[name] = value
        self.update_mod()

    def __getitem__(self, name):
        return self.cache[name]

    def __contains__(self, name):
        return name in self.cache

    def __delitem__(self, name):
        if name in self.cache:
            del self.cache[name]
        self.update_mod()

    def update_umod(self):
        if "control" not in self.cache:
            self.cache["control"] = self.cache['control']

        ts = int(time.mktime(time.gmtime()))
        self.cache["control"]["canto-user-modified"] = ts
        self.cache["control"]["canto-modified"] = ts

    def update_mod(self):
        if "control" not in self.cache:
            self.cache["control"] = self.cache['control']

        ts = int(time.mktime(time.gmtime()))
        self.cache["control"]["canto-modified"] = ts

    @wlock_feeds
    def sync(self):

        # If we get a sync after we're closed, or before we're open
        # just ignore it.

        if self.cache == {}:
            return

        f, tmpname = tempfile.mkstemp("", "feeds", os.path.dirname(self.filename))
        os.close(f)

        fp = gzip.open(tmpname, "wt", 9, "UTF-8")
        json.dump(self.cache, fp, indent=4, sort_keys=True)
        fp.close()

        log.debug("Written tempfile.")

        shutil.move(tmpname, self.filename)

        log.debug("Synced.")

    def close(self):
        log.debug("Closing.")
        self.sync()
        self.cache = {}
        call_hook("daemon_db_close", [self.filename])



# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .rwlock import RWLock

# NOTE: feed_lock and tag_lock only protect the existence of Feed() and Tag()
# objects, and their configuration. The Tag() objects have their own locks the
# protect their content.

feed_lock = RWLock('feed_lock')
tag_lock = RWLock('tag_lock')

# NOTE: if config_lock is held writable, feed_lock and tag_lock must also be
# held writable.

config_lock = RWLock('config_lock')

# The rest of these are independent.
watch_lock = RWLock('watch_lock')
attr_lock = RWLock('attr_lock')
socktran_lock = RWLock('socktran_lock')
hook_lock = RWLock('hook_look')



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import locale

locale_enc = locale.getpreferredencoding()

# These are basically just wrappers to close the encoding
# options into their scopes so we don't have to pass the 
# encoding around or call the above a million times.

def get_encoder(errors = "replace", encoding = None):
    if not encoding:
        encoding = locale_enc

    def encoder(s):
        return s.encode(encoding, errors)
    return encoder

encoder = get_encoder()



# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .plugins import PluginHandler, Plugin, try_plugins, set_program
from .client import CantoClient
from .format import escsplit
from .hooks import call_hook

from xml.sax.saxutils import escape as xml_escape
import xml.parsers.expat
import feedparser
import traceback
import pprint
import time
import sys

import logging

def assign_to_dict(d, var, val):
    terms = escsplit(var, '.', 0, 0, True)
    cur = d

    for term in terms[:-1]:
        if term not in cur:
            cur[term] = {}
        elif type(cur[term]) != dict:
            return (False, "Term %s is not dict" % term)
        cur = cur[term]

    cur[terms[-1]] = val
    return (True, val)

def access_dict(d, var):
    terms = escsplit(var, '.', 0, 0, True)
    cur = d

    for term in terms[:-1]:
        if term not in cur:
            return (False, False)
        elif type(cur) != dict:
            return (False, False)
        cur = cur[term]

    if terms[-1] not in cur:
        return (False, False)
    return (True, cur[terms[-1]])

class DaemonRemotePlugin(Plugin):
    pass

class CantoRemote(PluginHandler, CantoClient):
    def __init__(self):
        self.plugin_attrs = {}

        # By default this will log to stderr.
        logging.basicConfig(
                format = "%(asctime)s : %(name)s -> %(message)s",
                datefmt = "%H:%M:%S",
                level = logging.ERROR
        )

        version = "canto-remote " + "0.9.8" + " " + ""
        optl = self.common_args("h", ["help"], version)
        if optl == -1:
            sys.exit(-1)

        if self.args(optl) == -1:
            sys.exit(-1)

        set_program("canto-remote")
        try_plugins(self.conf_dir, self.plugin_default, self.disabled_plugins,
                self.enabled_plugins)

        PluginHandler.__init__(self)
        self.plugin_class = DaemonRemotePlugin
        self.update_plugin_lookups()

        try:
            if self.port < 0:
                self.start_daemon()
                CantoClient.__init__(self, self.socket_path)
            else:
                CantoClient.__init__(self, None,\
                        port = self.port, address = self.addr)
        except Exception as e:
            print("Error: %s" % e)
            print(self.socket_path)
            sys.exit(-1)

        self.handle_args()

    def args(self, optlist):
        for opt, arg in optlist:
            if opt in [ "-h", "--help" ]:
                self.print_help()
                sys.exit(0)
        return 0

    def print_help(self):
        print("USAGE: canto-remote [options] ([command] [command-args])\n")
        self.print_commands()
        print("\n\t-h/--help\tThis help")
        print("\t-V/--version\tPrint version")
        print("\t-v/\t\tVerbose logging (for debug)")
        print("\t-D/--dir <dir>\tSet configuration directory.")
        print("\nPlugin control\n")
        print("\t--noplugins\t\t\t\tDisable plugins")
        print("\t--enableplugins 'plugin1 plugin2...'\tEnable single plugins (overrides --noplugins)")
        print("\t--disableplugins 'plugin1 plugin2...'\tDisable single plugins")
        print("\nNetwork control\n")
        print("NOTE: These should be used in conjunction with SSH port forwarding to be secure\n")
        print("\t-a/--address <IP>\tConnect to this address")
        print("\t-p/--port <port>\tConnect to this port")

    def print_commands(self):
        print("COMMANDS")
        print("\thelp - get help on a command")
        print("\taddfeed - subscribe to a new feed")
        print("\tlistfeeds - list all subscribed feeds")
        print("\tdelfeed - unsubscribe from a feed")
        print("\tstatus - print item counts")
        print("\tforce-update - refetch all feeds")
        print("\tconfig - change / query configuration variables")
        print("\tone-config - change / query one configuration variable")
        print("\texport - export feed list as OPML")
        print("\timport - import feed list from OPML")
        print("\tkill - cleanly kill the daemon")
        print("\tscript - run script")
        call_hook("remote_print_commands", [])

    def _wait_response(self, cmd):
        r = None
        while True:
            r = self.read()
            if type(r) == int:
                if r == 16:
                    print("Server hung up.")
                else:
                    print("Got code: %d" % r)
                print("Please check daemon-log for exception.")
                return
            elif type(r) == tuple:
                if not cmd:
                    return r
                if r[0] == cmd:
                    return r[1]
                elif r[0] == "ERRORS":
                    print("ERRORS!")
                    for key in list(r[1].keys()):
                        for val, err in r[1][key]:
                            print("%s -> %s: %s" % (key, val, err))
            elif r:
                print("Unknown return: %s" % r)
                break
        return None

    def _autoname(self, URL):
        extra_headers = { 'User-Agent' :\
                'Canto/0.9.0 + http://codezen.org/canto-ng' }
        try:
            content = feedparser.parse(URL, request_headers = extra_headers)
        except Exception as e:
            print("ERROR: Couldn't determine name: %s" % e)
            return None

        if "title" in content["feed"]:
            return content["feed"]["title"]
        else:
            print("Couldn't find title in feed!")

        return None

    def _get_feeds(self):
        self.write("CONFIGS", [ "feeds" ])
        c = self._wait_response("CONFIGS")

        return c["feeds"]

    def _addfeed(self, attrs):
        # Fill out
        if "name" not in attrs or not attrs["name"]:
            attrs["name"] = self._autoname(attrs["url"])
        if not attrs["name"]:
            print("Failed to autoname, please specify!")
            return False

        print("Adding feed %s - %s" % (attrs["url"], attrs["name"]))

        # SET merges the config options, so f will be appended to the
        # current value of "feeds", rather than overwriting.

        self.write("SETCONFIGS", { "feeds" : [ attrs ] } )
        self.write("PING", [])
        self._wait_response("PONG")

        return True

    def cmd_addfeed(self):
        """USAGE: canto-remote addfeed [URL] (option=value) ...
    Where URL is the feed's URL. You can also specify options for the feed:

        name = Feed name (if not specified remote will attempt to lookup)
        rate = Rate, in minutes, at which this feed should be fetched.

        username = Username (if necessary) for password protected feeds.
        password = Password for password protected feeds."""

        if len(sys.argv) < 2:
            return False

        feed = { "url" : sys.argv[1] }
        name = None

        # Grab any feedopts from the commandline.

        for arg in sys.argv[2:]:
            opt, val = escsplit(arg, "=", 1, 1, True)
            if not opt or not val:
                print("ERROR: can't parse '%s' as x=y setting." % arg)
                continue
            feed[opt] = val

        return self._addfeed(feed)

    def cmd_listfeeds(self):
        """USAGE: canto-remote listfeeds
    Lists all tracked feeds."""

        if len(sys.argv) > 1:
            return False

        for idx, f in enumerate(self._get_feeds()):
            s = ("%d. " % idx) + f["name"] + " "

            if "alias" in f:
                s += "(" + f["alias"] + ")"

            s += "\n" + f["url"] + "\n"
            print(s)

    def cmd_delfeed(self):
        """USAGE: canto-remote delfeed [URL|name|alias]
    Unsubscribe from a feed."""
        if len(sys.argv) != 2:
            return False

        term = sys.argv[1]

        for idx, f in enumerate(self._get_feeds()):
            matches = [ f["url"], f["name"], "%s" % idx]
            if "alias" in f:
                matches.append(f["alias"])

            if term in matches:
                print("Unsubscribing from %s" % f["url"])
                self.write("DELCONFIGS",  { "feeds" : [ f ] })

    def _config(self, args, evaled=False):
        sets = {}
        gets = []

        for arg in args:

            if "=" not in arg:
                gets.append(arg)
                continue

            var, val = escsplit(arg, "=", 1, 1)
            var = var.lstrip().rstrip()

            # We'll want to read back any value, regardless
            gets.append(var)

            if evaled:
                try:
                    val = eval(val, {},{})
                    val_ok = True
                except Exception as e:
                    print("Unable to eval value: %s - %s" % (val, e))
                    val_ok = False
            else:
                val_ok = True

            if val_ok:
                val_ok, ret = assign_to_dict(sets, var, val)
                if not val_ok:
                    print(ret)

        if sets:
            self.write("SETCONFIGS", sets)

        self.write("CONFIGS", [])
        c = self._wait_response("CONFIGS")

        for var in gets:
            valid, value = access_dict(c, var)
            if valid:
                print("%s = %s" % (var, value))
            else:
                print("Couldn't get %s!" % var)

        return True

    def cmd_one_config(self):
        """USAGE: canto-remote one-config [--eval] [option] ( = value)
    Where option is a full variable declaration like 'section.variable' and
    value is any string. If value is omitted, the current value will be printed.

    This differs from config as only one option can be set/got at a time, but
    it allows lax argument parsing (i.e. one-config CantoCurses.browser =
    firefox will work as expected, without quoting.)

    If the value you're setting is a type other than string, you must specify
    --eval and the value will be eval'd into a proper type.

    NOTE: validation is done by the client that uses the variable, canto-remote
    will let you give bad values, or even set values to non-existent
    variables."""

        if len(sys.argv) < 2:
            return False

        if sys.argv[1] == "--eval":
            return self._config([" ".join(sys.argv[2:])], True)
        return self._config([" ".join(sys.argv[1:])], False)

    def cmd_config(self):
        """USAGE: canto-remote config [--eval] [option](=value) ...
    Where option is a full variable declaration like 'section.variable' and
    value is any string. If value is omitted, the current value will be printed.

    This differs from one-config as multiple sets/gets can be done, but it is
    more strict in terms of argument parsing.

    If the value you're setting is a type other than string, you must specify
    --eval and the value will be eval'd into a proper type.

    NOTE: validation is done by the client that uses the variable, canto-remote
    will let you give bad values, or even set values to non-existent
    variables."""

        if len(sys.argv) < 2:
            return False

        if sys.argv[1] == "--eval":
            return self._config(sys.argv[2:], True)
        return self._config(sys.argv[1:], False)

    def cmd_export(self):
        """USAGE: canto-remote export

    This will print an OPML file to standard output."""

        print("""<opml version="1.0">""")
        print("""\t<body>""")
        for f in self._get_feeds():
            print("""\t\t<outline text="%s" xmlUrl="%s" type="rss" />""" %\
                (xml_escape(f["name"].replace("\"","\\\"")),
                 xml_escape(f["url"])))

        print("""\t</body>""")
        print("""</opml>""")

    def cmd_import(self):
        """USAGE: canto-remote import [OPML file]

    This will automatically import feeds from an OPML file, which can be
    generated by many different feed readers and other programs."""

        if len(sys.argv) != 2:
            return False

        opmlpath = sys.argv[1]

        try:
            data = open(opmlpath, "r").read()
        except Exception as e:
            print("Couldn't read OPML file:")
            traceback.print_exc()
            return

        feeds = []
        def parse_opml(name, attrs):
            # Skip elements we don't care about.
            if name != "outline":
                return

            # Skip outline elements with unknown type.
            if "type" in attrs and attrs["type"] not in ["pie","rss"]:
                return

            # Skip outline elements with type, but no URL
            if "xmlUrl" not in attrs:
                return

            f = { "url" : attrs["xmlUrl"], "name" : None }
            if "text" in attrs:
                f["name"] = attrs["text"]

            feeds.append(f)

        parser = xml.parsers.expat.ParserCreate()
        parser.StartElementHandler = parse_opml
        parser.Parse(data.encode("UTF-8"), 1)

        for feed in feeds:
            self._addfeed(feed)

    def cmd_script(self):
        """USAGE canto-remote script (scriptfile)

    Run script from scriptfile or stdin.

    Note: This is intended for testing and does not gracefully handle errors."""

        if len(sys.argv) not in [1, 2]:
            return False

        if len(sys.argv) == 1:
            lines = sys.stdin.readlines()
        else:
            f = open(sys.argv[1], "r")
            lines = f.readlines()
            f.close()

        pp = pprint.PrettyPrinter()

        for line in lines:
            line = line[:-1].lstrip()
            print(line)
            sys.__stdout__.flush()

            # Wait for n responses.

            if line.startswith("REMOTE_WAIT "):
                num = int(line.split(" ", 1)[-1])
                for i in range(num):
                    r = self._wait_response(None)
                    print(pp.pformat(r))
                    sys.__stdout__.flush()

            elif line.startswith("REMOTE_IGNORE "):
                num = int(line.split(" ", 1)[-1])
                for i in range(num):
                    self._wait_response(None)

            # Hang with socket open so that the daemon thinks
            # we're using any data we've requested. Script runners
            # must be smart enough to signal-kill this remote.

            elif line.startswith("REMOTE_HANG"):
                while True:
                    time.sleep(1000)

            # Skip comments / blank

            elif line == '' or line.startswith("#"):
                continue

            else:
                cmd, arg = line.split(' ', 1)
                self.write(cmd, eval(arg))

    def cmd_kill(self):
        """USAGE: canto-remote kill

    Cleanly kill the connected daemon."""

        self.write("DIE", {})

    def cmd_force_update(self):
        """USAGE: canto-remote force-update

    Force fetch of all feeds."""

        self.write("FORCEUPDATE", {})

    def _numstate(self, tag, state):
        self.write("AUTOATTR", [ "canto-state" ])
        self.write("ITEMS", [ tag ])

        items = []

        cmd, r = self._wait_response(None)
        while cmd != "ITEMSDONE":
            if cmd == "ITEMS":
                items.extend(r[tag])
            cmd, r = self._wait_response(None)

        if len(items) == 0:
            return 0

        attrs = self._wait_response("ATTRIBUTES")

        if state == "unread":
            return len([ x for x in items if "read" not in attrs[x]["canto-state"]])
        elif state == "read":
            return len([ x for x in items if "read" in attrs[x]["canto-state"]])
        else:
            return len(items)

    def cmd_status(self):
        """USAGE: canto-remote status (--tag=tag) (--read|--total|--tags)

    Print the number of unread items tracked by the daemon.

    --tags can be given to print a summary for all existing tags

    --tag can be given to print the number of items in that particular tag
    (i.e. "maintag:Slashdot", "user:cool")

    --read can be given to print the number of read items
    --total can be given to print the total number of items

    NOTE: This is still subject to defaults.global_transform, and defined tag
    transforms, so if you're using filter_read, for example, it will never
    return a single read item (so --total will have no effect and --read will
    always return 0)."""

        state = "unread"
        if "--read" in sys.argv:
            state = "read"
        if "--total" in sys.argv:
            state = "all"

        self.write("LISTTAGS","")
        t = self._wait_response("LISTTAGS")

        if "--tags" in sys.argv:
            for tag in t:
                print("%s : %s" % (tag, self._numstate(tag, state)))
        elif "--tag" in sys.argv:
            if "--tag" == sys.argv[-1]:
                print("--tag must be followed by a tag name")
                sys.exit(-1)

            tag = sys.argv[sys.argv.index("--tag") + 1]
            if tag not in t:
                print("Unknown tag %s - use --tags to list known tags" % tag)
                sys.exit(-1)

            print("%s : %s" % (tag, self._numstate(tag, state)))
        else:
            print("%s" % sum([self._numstate(tag, state) for tag in t if tag.startswith("maintag:")]))

    def cmd_help(self):
        """USAGE: canto-remote help [command]"""
        if len(sys.argv) < 2:
            return False

        command = "cmd_" + sys.argv[1].replace("-","_")

        if hasattr(self, command):
            print(getattr(self, command).__doc__)
        else:
            print(self.cmd_help.__doc__)
            self.print_commands()

    def handle_args(self):
        if len(sys.argv) < 1:
            self.print_help()
            return

        command = "cmd_" + sys.argv[0].replace("-","_")

        if hasattr(self, command):
            func = getattr(self, command)
            r = func()
            if r == False:
                print(func.__doc__)
        else:
            self.print_help()



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .protocol import CantoSocket
from .hooks import call_hook

from socket import SHUT_RDWR
from threading import Thread, Lock
import traceback
import logging
import select

log = logging.getLogger("SERVER")

class CantoServer(CantoSocket):
    def __init__(self, socket_name, dispatch, **kwargs):
        kwargs["server"] = True
        CantoSocket.__init__(self, socket_name, **kwargs)
        self.dispatch = dispatch
        self.conn_thread = None

        self.connections_lock = Lock()
        self.connections = [] # (socket, thread) tuples
        self.alive = True

        self.start_conn_loop()

    # Endlessly consume data from the connection. If there's enough data
    # for a complete command, toss it on the shared Queue.Queue

    def read_loop(self, conn):
        try:
            while self.alive:
                d = self.do_read(conn)
                if d:
                    if d == select.POLLHUP:
                        log.info("Connection ended.")
                        return
                    self.dispatch(conn, d)
        except Exception as e:
            tb = traceback.format_exc()
            log.error("Response thread dead on exception:")
            log.error("\n" + "".join(tb))
            return

    # Sit and select for connections on sockets:

    def conn_loop(self, sockets):
        while self.alive:
            try:
                # select with a timeout so we can check we're still alive
                r, w, x = select.select(sockets, [], sockets, 1)
                for s in sockets:
                    # If socket is readable, it's got a pending connection.
                    if s in r:
                        conn = s.accept()
                        log.info("conn %s from sock %s" % (conn, s))
                        self.accept_conn(conn[0])
            except Exception as e:
                tb = traceback.format_exc()
                log.error("Connection monitor exception:")
                log.error("\n" + "".join(tb))
                log.error("Attempting to continue.")

    def start_conn_loop(self):
        self.conn_thread = Thread(target = self.conn_loop,
                args = (self.sockets,), name = "Connection Monitor")
        self.conn_thread.daemon = True
        self.conn_thread.start()
        log.debug("Spawned connection monitor thread.")

    # Remove dead connection threads.

    def no_dead_conns(self):
        self.connections_lock.acquire()
        for c, t in self.connections[:]:
            if not t.is_alive():
                call_hook("server_kill_socket", [c])
                t.join()
                c.close()
                self.connections.remove((c, t))
                if self.connections == []:
                    call_hook("server_no_connections", [])
        self.connections_lock.release()

    def accept_conn(self, conn):
        self.read_locks[conn] = Lock()
        self.write_locks[conn] = Lock()
        self.write_frags[conn] = None

        # Notify watchers about new socket.
        call_hook("server_new_socket", [conn])

        self.connections_lock.acquire()

        self.connections.append((conn,\
                Thread(target = self.read_loop,\
                       args = (conn,), name="Connection #%s" %\
                       (len(self.connections)))
                ))

        self.connections[-1][1].daemon = True
        self.connections[-1][1].start()

        if len(self.connections) == 1:
            call_hook("server_first_connection", [])

        self.connections_lock.release()

        log.debug("Spawned new thread.")

    # Write a (cmd, args) to a single connection.
    def write(self, conn, cmd, args):
        if not conn:
            return None
        return self.do_write(conn, cmd, args)

    # Write a (cmd, args) to every connection.
    def write_all(self, cmd, args):
        self.no_dead_conns()

        self.connections_lock.acquire()
        for conn, t in self.connections:
            self.do_write(conn, cmd, args)
        self.connections_lock.release()

    def exit(self):
        self.alive = False
        self.conn_thread.join()

        # No locking, as we should already be single-threaded

        for conn, t in self.connections:
            conn.shutdown(SHUT_RDWR)
            conn.close()



#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .hooks import on_hook, call_hook
from .rwlock import read_lock, write_lock
from .locks import *

import logging

log = logging.getLogger("TAG")

class CantoTags():
    def __init__(self):
        self.tags = {}
        self.changed_tags = []

        # Per-tag transforms
        self.tag_transforms = {}

        # Extra tag map
        # This allows tags to be defined as parts of larger tags.  For example,
        # Penny Arcade and xkcd could both have the extra "comic" tag which
        # could then be used in a filter to implement categories.

        self.extra_tags = {}

    def items_to_tags(self, ids):
        tags = []
        for id in ids:
            for tag in self.tags:
                if id in self.tags[tag] and tag not in tags:
                    tags.append(tag)
        return tags

    def tag_changed(self, tag):
        if tag not in self.changed_tags:
            self.changed_tags.append(tag)

    def get_tag(self, tag):
        if tag in list(self.tags.keys()):
            return self.tags[tag]
        return []

    def get_tags(self):
        return list(self.tags.keys())

    def tag_transform(self, tag, transform):
        self.tag_transforms[tag] = transform

    def set_extra_tags(self, tag, extra_tags):
        self.extra_tags[tag] = extra_tags

    def clear_tags(self):
        self.tags = {}

    def reset(self):
        self.tag_transforms = {}
        self.extra_tags = {}

        # Don't set tag_changed here, because we don't want to alert when a tag
        # will probably just be re-populated with identical content.

        # It it isn't, then the add or remove will set it for us.

        self.clear_tags()

    #
    # Following must be called with tag_lock held with write
    #

    def add_tag(self, id, name):
        if name in self.extra_tags:
            extras = self.extra_tags[name]
        else:
            extras = []

        alladded = [ name ] + extras

        for name in alladded:
            # Create tag if no tag exists
            if name not in self.tags:
                self.tags[name] = []
                call_hook("daemon_new_tag", [[ name ]])

            # Add to tag.
            if id not in self.tags[name]:
                self.tags[name].append(id)
                self.tag_changed(name)

    def remove_tag(self, id, name):
        if name in self.tags and id in self.tags[name]:
            self.tags[name].remove(id)
            self.tag_changed(name)

    def remove_id(self, id):
        for tag in self.tags:
            if id in self.tags[tag]:
                self.tags[tag].remove(id)
                self.tag_changed(tag)

    def apply_transforms(self, tag, tagobj):
        from .config import config
        # Global transform
        if config.global_transform:
            tagobj = config.global_transform(tagobj)

        # Tag level transform
        if tag in self.tag_transforms and\
                self.tag_transforms[tag]:
            tagobj = self.tag_transforms[tag](tagobj)

        return tagobj

    def do_tag_changes(self):
        for tag in self.changed_tags:
            tagobj = self.get_tag(tag)

            try:
                tagobj = self.apply_transforms(tag, tagobj)
            except Exception as e:
                log.error("Exception applying transforms: %s" % e)

            self.tags[tag] = tagobj
            call_hook("daemon_tag_change", [ tag ])
        self.changed_tags = []

alltags = CantoTags()



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import traceback
import logging
import sys
import os

log = logging.getLogger("PLUGINS")

class CantoWrongProgramException(Exception):
    pass

PROGRAM="unset"

def set_program(program_name):
    global PROGRAM
    PROGRAM=program_name

def check_program(*args):
    global PROGRAM
    if PROGRAM not in args:
        raise CantoWrongProgramException

def try_plugins(topdir, plugin_default=True, disabled_plugins=[], enabled_plugins=[]):
    p = topdir + "/plugins"
    pinit = p + "/__init__.py"

    if not os.path.exists(p):
        log.info("Creating plugins directory.")
        try:
            os.mkdir(p)
        except Exception as e:
            tb = traceback.format_exc()
            log.error("Exception creating plugin directory")
            log.error("\n" + "".join(tb))
            return
    elif not os.path.isdir(p):
        log.warn("Plugins file is not directory.")
        return

    if not os.path.exists(pinit):
        log.info("Creating plugin __init__.py")
        try:
            f = open(pinit, "w")
            f.close()
        except Exception as e:
            tb = traceback.format_exc()
            log.error("Exception creating plugin __init__.py")
            log.error("\n" + "".join(tb))
            return

    # Add plugin path to front of Python path.
    sys.path.insert(0, topdir)

    all_errors = ""

    # Go ahead and import all .py
    for fname in sorted(os.listdir(p)):
        if fname.endswith(".py") and fname != "__init__.py":
            try:
                proper = fname[:-3]

                if plugin_default:
                    if proper in disabled_plugins:
                        log.info("[plugin] %s - DISABLED" % proper)
                    else:
                        __import__("plugins." + proper)
                        log.info("[plugin] %s" % proper)
                else:
                    if proper in enabled_plugins:
                        __import__("plugins." + proper)
                        log.info("[plugin] %s - ENABLED" % proper)
                    else:
                        log.info("[plugin] %s - DISABLED" % proper)
            except CantoWrongProgramException:
                pass
            except Exception as e:
                tb = traceback.format_exc()
                log.error("Exception importing file %s" % fname)
                nice = "".join(tb)
                all_errors += nice
                log.error(nice)

    if all_errors != "":
        return all_errors

class PluginHandler(object):
    def __init__(self):
        self.plugin_attrs = {}

    def update_plugin_lookups(self):
        # Populate a dict of overridden attributes

        self.plugin_attrs = {}

        self.plugin_class_instances =\
                [ c(self) for c in self.plugin_class.__subclasses__() ]

        for iclass in self.plugin_class_instances[:]:
            try:
                # Warn if we're overriding a previously defined plugin attr
                for iclass_attr in list(iclass.plugin_attrs.keys()):
                    if iclass_attr in self.plugin_attrs:
                        log.warn("Multiply defined plugin attribute!: %s" %\
                                iclass_attr)

                self.plugin_attrs.update(iclass.plugin_attrs)
            except Exception as e:
                log.error("Error initializing plugins:")
                log.error(traceback.format_exc())

                # Malformed plugins removed from instances
                self.plugin_class_instances.remove(iclass)
                continue

    def __getattribute__(self, name):
        if name == "plugin_attrs" or name not in self.plugin_attrs:
            return object.__getattribute__(self, name)
        return self.plugin_attrs[name]

# Plugin is the base class for all of the separate plugin classes for each Gui
# object. There are two reasons to pin plugins to an empty class:
#
# - 'object' in the hierarchy via PluginHandler means we can use
#   __subclasses__, the cornerstone of the plugins system
#
# - This allows the plugins to have a hard distinction between self (the
#   instantiated class object) and obj (the instantiated main object that's
#   being overridden). This means that plugins don't have to worry about
#   clobbering anything.
#
# As a side effect, using the separate plugin architecture, we also can
# enable/disable pluggability on a class basis. For example, if TagList
# didn't specify a plugin_class, then it could not be overridden or hooked.

class Plugin(object):
        pass



# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

# This Backend class is the core of the daemon's specific protocol.

CANTO_PROTOCOL_VERSION = 0.9

from .feed import allfeeds, wlock_all, stop_feeds, rlock_feed_objs, runlock_feed_objs
from .encoding import encoder
from .server import CantoServer
from .config import config, parse_locks, parse_unlocks
from .storage import CantoShelf
from .fetch import CantoFetch
from .hooks import on_hook, call_hook
from .tag import alltags
from .transform import eval_transform
from .plugins import PluginHandler, Plugin, try_plugins, set_program
from .rwlock import alllocks, write_lock, read_lock
from .locks import *

import traceback
import logging
import signal
import fcntl
import errno
import time
import sys
import os

# By default this will log to stderr.
logging.basicConfig(
        format = "%(asctime)s : %(name)s -> %(message)s",
        datefmt = "%H:%M:%S",
        level = logging.INFO
)

log = logging.getLogger("CANTO-DAEMON")

class DaemonBackendPlugin(Plugin):
    pass

# Index threads and the main thread no longer take multiple locks at once. The
# cmd_* functions in CantoBackend only need to worry about deadlocking with
# each other.

class CantoBackend(PluginHandler, CantoServer):
    def __init__(self):

        # Nothing referenced before try_plugins should be
        # pluggable.

        self.plugin_attrs = {}

        # Shelf for feeds:
        self.fetch = None
        self.fetch_manual = False
        self.fetch_force = False

        # Whether fetching is inhibited.
        self.no_fetch = False

        self.watches = { "new_tags" : [],
                         "del_tags" : [],
                         "config" : [],
                         "tags" : {} }

        self.autoattr = {}

        # Per socket transforms.
        self.socket_transforms = {}

        self.shelf = None

        # No bad arguments.
        version = "canto-daemon " + "0.9.8" + " " + ""
        optl = self.common_args("nhc:",["nofetch","help","cache="], version)
        if optl == -1:
            sys.exit(-1)

        if self.args(optl):
            sys.exit(-1)

        # No invalid paths.
        if self.ensure_paths():
            sys.exit(-1)

        # Get pid lock.
        if self.pid_lock():
            sys.exit(-1)

        # Previous to this line, all output is just error messages to stderr.
        self.set_log()

        # Initial log chatter.
        log.info(version)

        if self.verbosity:
            rootlog = logging.getLogger()
            rootlog.setLevel(max(rootlog.level - 10 * self.verbosity,0))
            log.info("verbosity = %d" % self.verbosity)

        log.info("conf_dir = %s" % self.conf_dir)

        # Evaluate any plugins
        set_program("canto-daemon")
        try_plugins(self.conf_dir, self.plugin_default, self.disabled_plugins,
                self.enabled_plugins)

        PluginHandler.__init__(self)

        self.plugin_class = DaemonBackendPlugin
        self.update_plugin_lookups()

        if self.no_fetch:
            log.info("NOFETCH, will not be automatically updating.")

        # Actual start.
        self.get_storage()

        self.get_config()

        self.get_fetch()

        self.setup_hooks()

        self.sfile = self.conf_dir + "/.canto_socket"
        log.info("Listening on unix socket: %s" % self.sfile)

        try:
            if self.port < 0:
                CantoServer.__init__(self, self.sfile, self.socket_command)
            else:
                log.info("Listening on interface %s:%d" %\
                        (self.addr, self.port))
                CantoServer.__init__(self, self.sfile, self.socket_command,\
                        port = self.port, interface = self.addr)
        except Exception as e:
            err = "Error: %s" % e
            print(err)
            log.error(err)
            call_hook("daemon_exit", [])
            sys.exit(-1)

        # Signal handlers kickoff after everything else is init'd

        self.interrupted = 0

        signal.signal(signal.SIGINT, self.sig_int)
        signal.signal(signal.SIGTERM, self.sig_int)
        signal.signal(signal.SIGUSR1, self.sig_usr)

        self.start()

    def on_config_change(self, change, originating_socket):

        config.parse(False, change)

        log.debug("config.errors = %s", config.errors)

        if config.errors:
            self.write(originating_socket, "ERRORS", config.errors)
            config.parse()

            # No changes actually realized, bail
            return
        else:
            config.write()

        # Kill feeds that haven't been re-instantiated.
        allfeeds.all_parsed()

        # Force check of fetching. This automatically starts the fetch. For new
        # feeds, but also takes any new settings (like rates) into account.

        self.fetch_force = True

        # Pretend that the sockets *other* than the ones that made the change
        # issued a CONFIGS for each of the root keys.

        for socket in self.watches["config"]:
            # Don't echo changes back to socket that made them.
            if socket != originating_socket:
                self.in_configs(list(change.keys()), socket)

    # Notify clients of new tags.

    @read_lock(watch_lock)
    def on_new_tag(self, tags):
        for socket in self.watches["new_tags"]:
            self.write(socket, "NEWTAGS", tags)

    # Propagate tag changes to watching sockets.

    @read_lock(watch_lock)
    def on_tag_change(self, tag):
        if tag in self.watches["tags"]:
            for socket in self.watches["tags"][tag]:
                self.write(socket, "TAGCHANGE", tag)

    # Notify clients of dead tags:

    @read_lock(watch_lock)
    def on_del_tag(self, tags):
        for socket in self.watches["del_tags"]:
            self.write(socket, "DELTAGS", tags)

    # If a socket dies, it's no longer watching any events.

    @write_lock(socktran_lock)
    @write_lock(watch_lock)
    def on_kill_socket(self, socket):
        while socket in self.watches["config"]:
            self.watches["config"].remove(socket)

        while socket in self.watches["new_tags"]:
            self.watches["new_tags"].remove(socket)

        while socket in self.watches["del_tags"]:
            self.watches["del_tags"].remove(socket)

        for tag in self.watches["tags"]:
            while socket in self.watches["tags"][tag]:
                self.watches["tags"][tag].remove(socket)

        if socket in list(self.socket_transforms.keys()):
            del self.socket_transforms[socket]

    # We need to be alerted on certain events, ensure
    # we get notified about them.

    def setup_hooks(self):
        on_hook("daemon_new_tag", self.on_new_tag)
        on_hook("daemon_del_tag", self.on_del_tag)
        on_hook("daemon_config_change", self.on_config_change)
        on_hook("daemon_tag_change", self.on_tag_change)
        on_hook("server_kill_socket", self.on_kill_socket)

        # For plugins
        on_hook("daemon_set_configs", lambda x, y : self.internal_command(x, self.in_setconfigs, y))
        on_hook("daemon_del_configs", lambda x, y : self.internal_command(x, self.in_delconfigs, y))
        on_hook("daemon_get_configs", lambda x, y : self.internal_command(x, self.in_configs, y))

    # VERSION -> X.Y

    def cmd_version(self, socket, args):
        self.write(socket, "VERSION", CANTO_PROTOCOL_VERSION)

    # PING -> PONG

    def cmd_ping(self, socket, args):
        self.write(socket, "PONG", "")

    # LISTTAGS -> [ "tag1", "tag2", .. ]
    # This makes no guarantee on order *other* than the fact that
    # maintag tags will be first, and in feed order. Following tags
    # are in whatever order the dict gives them in.

    @read_lock(feed_lock)
    @read_lock(tag_lock)
    def cmd_listtags(self, socket, args):
        r = []
        for feed in allfeeds.get_feeds():
            r.append("maintag:" + feed.name)
        for tag in alltags.get_tags():
            if tag not in r:
                r.append(tag)

        self.write(socket, "LISTTAGS", r)

    # LISTTRANSFORMS -> [ { "name" : " " } for all defined filters ]

    @read_lock(config_lock)
    def cmd_listtransforms(self, socket, args):
        transforms = []
        for transform in config.transforms:
            transforms.append({"name" : transform["name"]})
        self.write(socket, "LISTTRANSFORMS", transforms)

    # TRANSFORM {} -> return current socket transform, with names instead of
    # actual filt objects.
    # TRANSFORM {"string":"transform"} -> set a socket transform
    # TRANSFORM {"string": None } -> un set a socket transform

    @write_lock(socktran_lock)
    def cmd_transform(self, socket, args):
        if not args:
            if socket in self.socket_transforms:
                str_dict = {}
                for filt in self.socket_transforms[socket]:
                    str_dict[filt] = str(self.socket_transforms[socket][filt])
                self.write(socket, "TRANSFORM", str_dict)
            else:
                self.write(socket, "TRANSFORM", {})
            return

        if socket not in self.socket_transforms:
            self.socket_transforms[socket] = {}

        for key in args:
            # Unset beforehand means query.
            if not args[key]:
                if key in self.socket_transforms[socket]:
                    self.write(socket, "TRANSFORM", { key : str(self.socket_transforms[socket][key])})
                else:
                    self.write(socket, "TRANSFORM", { key : "None" })
                continue

            filt = None
            try:
                filt = eval_transform(args[key])
            except Exception as e:
                self.write(socket, "EXCEPT",\
                        "Couldn't parse transform: %s\n%s" % (args[key], e))
                continue

            if filt == None:
                if key in self.socket_transforms[socket]:
                    log.debug("Unsetting socket transform %s:%s", socket, key)
                    del self.socket_transforms[socket][key]
                continue

            log.debug("Setting socket transform: %s:%s = %s", socket, key, filt)
            self.socket_transforms[socket][key] = filt

    # AUTOATTR [ attrs ... ] -> Follow up each items request with
    # an attributes request for attrs.

    # This command is intended to reduce round trip time and allow
    # clients to become informative quickly by making the individual
    # story IDs unnecessary to request information about them.

    # Hold attr_lock just to keep cmd_item from trying to use autoattr
    @write_lock(attr_lock)
    def cmd_autoattr(self, socket, args):
        self.autoattr[socket] = args

    # ITEMS [tags] -> { tag : [ ids ], tag2 : ... }

    @read_lock(attr_lock)
    @read_lock(feed_lock)
    def _apply_socktrans(self, socket, tag):
        feeds = allfeeds.items_to_feeds(tag)
        rlock_feed_objs(feeds)
        socktran_lock.acquire_read()
        try:

            for filt in self.socket_transforms[socket]:
                tag = self.socket_transforms[socket][filt](tag)
        finally:
            socktran_lock.release_read()
            runlock_feed_objs(feeds)
        return tag

    def cmd_items(self, socket, args):
        ids = []
        response = {}

        for tag in args:
            items = alltags.get_tag(tag)

            if socket in self.socket_transforms:
                items = self._apply_socktrans(socket, items)

            attr_list = []

            if len(items) == 0:
                self.write(socket, "ITEMS", { tag : [] })
            else:
                attr_req = {}
                if socket in self.autoattr:
                    for id in items:
                        attr_req[id] = self.autoattr[socket][:]
                    attr_list.append(attr_req)

                self.write(socket, "ITEMS", { tag : items })

            self.write(socket, "ITEMSDONE", {})

            for attr_req in attr_list:
                self.cmd_attributes(socket, attr_req)

    # ATTRIBUTES { id : [ attribs .. ] .. } ->
    # { id : { attribute : value } ... }

    # Hold feed_lock so that get_attributes won't fail on a missing feed, but
    # items_to_feeds can still throw an exception if attributes requests come
    # in for items from removed feeds.

    @read_lock(feed_lock)
    def cmd_attributes(self, socket, args):
        ret = {}
        feeds = allfeeds.items_to_feeds(list(args.keys()))
        for f in feeds:
            ret.update(f.get_attributes(feeds[f], args))

        self.write(socket, "ATTRIBUTES", ret)

    # SETATTRIBUTES { id : { attribute : value } ... } -> None

    @read_lock(feed_lock)
    @write_lock(tag_lock)
    def cmd_setattributes(self, socket, args):

        feeds = allfeeds.items_to_feeds(list(args.keys()))
        for f in feeds:
            f.set_attributes(feeds[f], args)

        tags = alltags.items_to_tags(list(args.keys()))
        for t in tags:
            call_hook("daemon_tag_change", [ t ])

    # CONFIGS [ "top_sec", ... ] -> { "top_sec" : full_value }

    # Internally, called only by functions that hold read or write on
    # config_lock

    def in_configs(self, args, socket=None):
        if args:
            ret = {}
            for topsec in args:
                if topsec in config.json:
                    ret[topsec] = config.json[topsec]
        else:
            ret = config.json

        if socket:
            self.write(socket, "CONFIGS", ret)
        return ret

    # External, needs to grab lock.

    @read_lock(config_lock)
    def cmd_configs(self, socket, args):
        ret = self.in_configs(args, socket)

    # SETCONFIGS { "key" : "value", ...}

    def in_setconfigs(self, args):
        self.cmd_setconfigs(None, args)
        return config.json

    def cmd_setconfigs(self, socket, args):
        parse_locks()

        config.merge(args.copy())

        # config_change handles it's own locking
        call_hook("daemon_config_change", [args, socket])

        parse_unlocks()

    # DELCONFIGS { "key" : "DELETE", ...}

    def in_delconfigs(self, args):
        self.cmd_delconfigs(None, args)
        return config.json

    def cmd_delconfigs(self, socket, args):
        parse_locks()

        config.delete(args.copy())

        # config_change handles it's own locking
        call_hook("daemon_config_change", [args, socket])

        parse_unlocks()

    # WATCHCONFIGS

    @write_lock(watch_lock)
    def cmd_watchconfigs(self, socket, args):
        if socket not in self.watches["config"]:
            self.watches["config"].append(socket)

    # WATCHNEWTAGS

    @write_lock(watch_lock)
    def cmd_watchnewtags(self, socket, args):
        if socket not in self.watches["new_tags"]:
            self.watches["new_tags"].append(socket)

    # WATCHDELTAGS

    @write_lock(watch_lock)
    def cmd_watchdeltags(self, socket, args):
        if socket not in self.watches["del_tags"]:
            self.watches["del_tags"].append(socket)

    # WATCHTAGS [ "tag", ... ]

    @write_lock(watch_lock)
    def cmd_watchtags(self, socket, args):
        for tag in args:
            log.debug("socket %s watching tag %s", socket, tag)
            if tag in self.watches["tags"]:
                if socket not in self.watches["tags"][tag]:
                    self.watches["tags"][tag].append(socket)
            else:
                self.watches["tags"][tag] = [socket]

    # UPDATE {}

    # Note that this is intended to allow clients to take manual
    # control when canto is started with --nofetch and doesn't
    # override rates or any other factors in updating.

    def cmd_update(self, socket, args):
        self.fetch_manual = True
        self.fetch_force = False

    # FORCEUPDATE {}

    # This command, on the other hand, *will* force the timers.

    def cmd_forceupdate(self, socket, args):
        self.fetch_manual = True
        self.fetch_force = True

    # The workhorse that maps all requests to their handlers.

    def socket_command(self, socket, data):
        cmd, args = data

        if cmd == "DIE":
            log.info("Received DIE.")
            self.interrupted = True
        else:
            cmdf = "cmd_" + cmd.lower()
            if hasattr(self, cmdf):
                func = getattr(self, cmdf)

                call_hook("daemon_pre_" + cmd.lower(), [socket, args])

                try:
                    func(socket, args)
                except Exception as e:
                    tb = "".join(traceback.format_exc())
                    self.write(socket, "EXCEPT", tb)
                    log.error("Protocol exception:")
                    log.error("\n" + tb)

                call_hook("daemon_post_" + cmd.lower(), [socket, args])
            else:
                log.info("Got unknown command: %s" % (cmd))

    def internal_command(self, cb, func, args):
        r = func(args)
        if cb:
            cb(r)

    def run(self):

        # Start fetch threads to load from disk. No need for locking as we
        # haven't started any threads yet
        self.fetch.fetch(True, True)

        log.debug("Beginning to serve...")
        call_hook("daemon_serving", [])
        while 1:
            if self.interrupted:
                log.info("Interrupted. Exiting.")
                return

            # Clean up any dead connection threads.
            self.no_dead_conns()

            # Clean up any threads done updating.
            self.fetch.reap()

            # Check whether feeds need to be updated and fetch
            # them if necessary.

            if (not self.no_fetch or self.fetch_manual):
                self.fetch.fetch(self.fetch_force, False)

                self.fetch_manual = False
                self.fetch_force = False

            call_hook("daemon_end_loop", [])

            time.sleep(1)

    # Shutdown cleanly

    def cleanup(self):
        # Stop feeds, will cause feed.index() threads to bail without
        # Messing with the disk.

        stop_feeds()

        # Grab locks to keep any other write usage from happening.

        wlock_all()

        self.shelf.close()

        call_hook("daemon_exit", [])

        # The rest of this is bonus, the important part is to protect the disk.
        log.debug("DB shutdown.")

        # Force all connection threads to end.

        self.exit()

        # Wait for all fetches to end.

        self.fetch.reap(True)

        # Delete the socket file, so it can only be used when we're actually
        # listening.

        self.remove_socketfile()

        # Unlock the pidfile so another daemon could take over. Probably don't
        # have to do this since we're about to sys.exit anyway, but why not.

        self.pid_unlock()

        log.info("Exiting cleanly.")

    def print_help(self):
        print("USAGE: canto-daemon [options]")
        print("\t-h/--help\tThis help")
        print("\t-V/--version\tPrint version")
        print("\t-v/\t\tVerbose logging (for debug)")
        print("\t-D/--dir <dir>\tSet configuration directory.")
        print("\t-n/--nofetch\tJust serve content, don't fetch new content.")
        print("\n\nPlugin control\n")
        print("\t--noplugins\t\t\t\tDisable plugins")
        print("\t--enableplugins 'plugin1 plugin2...'\tEnable single plugins (overrides --noplugins)")
        print("\t--disableplugins 'plugin1 plugin2...'\tDisable single plugins")
        print("\nNetwork control\n")
        print("NOTE: These should be used in conjunction with SSH port forwarding to be secure\n")
        print("\t-a/--address <IP>\tBind to interface with this address")
        print("\t-p/--port <port>\tBind to this port")

    # This function parses and validates all of the command line arguments.
    def args(self, optlist):
        for opt, arg in optlist:
            if opt in ["-n", "--nofetch"]:
                self.no_fetch = True
            elif opt in ['-h', '--help']:
                self.print_help()
                sys.exit(0)
        return 0

    def sig_int(self, a, b):
        log.info("Received INT")
        self.interrupted = 1

    def sig_usr(self, a, b):
        import threading
        import gc

        held_locks = {}
        code = {}
        curthreads = threading.enumerate()

        for threadId, stack in sys._current_frames().items():
            name = str(threadId)
            for ct in curthreads:
                if ct.ident == threadId:
                    name = ct.name

            code[name] = ["NAME: %s" % name]
            for filename, lineno, fname, line in traceback.extract_stack(stack):
                code[name].append('FILE: "%s", line %d, in %s' % (filename, lineno, fname))
                if line:
                    code[name].append("  %s" % (line.strip()))

            held_locks[name] = ""
            for lock in alllocks:
                if lock.writer_id == threadId:
                    held_locks[name] += ("%s(w)" % lock.name)
                    continue
                for reader_id, reader_stack in lock.reader_stacks:
                    if reader_id == threadId:
                        held_locks[name] += ("%s(r)" % lock.name)

        for k in code:
            log.info('\n\nLOCKS: %s \n%s' % (held_locks[k], '\n'.join(code[k])))

        log.info("\n\nSTACKS:")
        for lock in alllocks:
            for (reader_id, reader_stack) in lock.reader_stacks:
                log.info("Lock %s (%s readers)" % (lock.name, lock.readers))
                log.info("Lock reader (thread %s):" % (reader_id,))
                log.info(''.join(reader_stack))

            for writer_stack in lock.writer_stacks:
                log.info("Lock %s (%s readers)" % (lock.name, lock.readers))
                log.info("Lock writer (thread %s):" % (lock.writer_id,))
                log.info(''.join(writer_stack))

        self.shelf.sync()
        gc.collect()

        # If we've got pympler installed, output a summary of memory usage.

        try:
            from pympler import summary, muppy
            summary.print_(summary.summarize(muppy.get_objects()))
        except:
            pass

    # This function makes sure that the configuration paths are all R/W or
    # creatable.

    def ensure_paths(self):
        if os.path.exists(self.conf_dir):
            if not os.path.isdir(self.conf_dir):
                log.error("Error: %s is not a directory." % self.conf_dir)
                return -1
            if not os.access(self.conf_dir, os.R_OK):
                log.error("Error: %s is not readable." % self.conf_dir)
                return -1
            if not os.access(self.conf_dir, os.W_OK):
                log.error("Error: %s is not writable." % self.conf_dir)
                return -1
        else:
            try:
                os.makedirs(self.conf_dir)
            except Exception as e:
                log.error("Exception making %s : %s" % (self.conf_dir, e))
                return -1
        return self.ensure_files()

    def ensure_files(self):
        for f in [ "feeds", "conf", "daemon-log", "pid"]:
            p = self.conf_dir + "/" + f
            if os.path.exists(p):
                if not os.path.isfile(p):
                    log.error("Error: %s is not a file." % p)
                    return -1
                if not os.access(p, os.R_OK):
                    log.error("Error: %s is not readable." % p)
                    return -1
                if not os.access(p, os.W_OK):
                    log.error("Error: %s is not writable." % p)
                    return -1

        # These paths are now guaranteed to read/writable.

        self.feed_path = self.conf_dir + "/feeds"
        self.pid_path = self.conf_dir + "/pid"
        self.log_path = self.conf_dir + "/daemon-log"
        self.conf_path = self.conf_dir + "/conf"

        return None

    def pid_lock(self):
        self.pidfile = open(self.pid_path, "a+")
        try:
            fcntl.flock(self.pidfile.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
            self.pidfile.seek(0, 0)
            self.pidfile.truncate()
            self.pidfile.write("%d" % os.getpid())
            self.pidfile.flush()
        except IOError as e:
            if e.errno == errno.EAGAIN:
                log.error("Error: Another canto-daemon is running here.")
                return -1
            raise
        return None

    def pid_unlock(self):
        log.debug("Unlocking pidfile.")
        fcntl.flock(self.pidfile.fileno(), fcntl.LOCK_UN)
        self.pidfile.close()
        log.debug("Unlocked.")

    # Reset basic log info to log to the right file.
    def set_log(self):
        f = open(self.log_path, "w")
        os.dup2(f.fileno(), sys.stderr.fileno())

    # Bring up storage, the only errors possible at this point are 
    # fatal and handled lower in CantoShelf.

    def get_storage(self):
        self.shelf = CantoShelf(self.feed_path)

    # Bring up config, the only errors possible at this point will
    # be fatal and handled lower in CantoConfig.

    def get_config(self):
        config.init(self.conf_path, self.shelf)
        config.parse()
        if config.errors:
            print("ERRORS:")
            for key in list(config.errors.keys()):
                for value, error in config.errors[key]:
                    s = "\t%s -> %s: %s" % (key, value, error)
                    print(encoder(s))

            sys.exit(-1)

    def get_fetch(self):
        self.fetch = CantoFetch(self.shelf)

    def remove_socketfile(self):
        os.unlink(self.sfile)

    def start(self):
        try:
            self.run()

        # Cleanly shutdown on ^C.
        except KeyboardInterrupt:
            pass

        # Pretty print any non-Keyboard exceptions.
        except Exception as e:
            tb = traceback.format_exc()
            log.error("Exiting on exception:")
            log.error("\n" + "".join(tb))

        self.cleanup()
        sys.exit(0)



# Canto Script Plugin
# by Jack Miller
# v1.0

# With this plugin you can add a feed with a URL starting with "script:" and
# ending with a simple script invocation. You must specify a name for the feed.
#
# For example:
#   canto-remote addfeed "script:~/bin/feed.sh" name="Script Feed"
#
# Scripts must have executable permissions.
#
# Scripts can be passed constant arguments, but are not executed in a shell
# environment.
#
# The script path should be absolute, or relative to home (~/), relative path
# behavior is undefined.
#
# Scripts are expected to output parsable RSS/Atom XML to stdout.

from canto_next.plugins import check_program

check_program("canto-daemon")

from canto_next.fetch import DaemonFetchThreadPlugin
from canto_next.feed import DaemonFeedPlugin

import feedparser
import subprocess
import logging
import shlex
import os

log = logging.getLogger("SCRIPT")

def debug(message):
    if EXTRA_LOG_OUTPUT:
        log.debug(message)

class ScriptFetch(DaemonFetchThreadPlugin):
    def __init__(self, fetch_thread):
        self.plugin_attrs = {
                "fetch_script" : self.fetch_script,
        }

    def fetch_script(self, **kwargs):
        if not kwargs["feed"].URL.startswith("script:"):
            return

        path = os.path.expanduser(kwargs["feed"].URL[7:])

        log.debug("path: %s", path)

        path = shlex.split(path)

        log.debug("split: %s", path)

        output = subprocess.check_output(path)

        log.debug("output: %s", output)

        contents = kwargs["newcontent"]
        contents.clear()

        feed = feedparser.parse(output)

        for key in feed:
            contents[key] = feed[key]



# -*- coding: utf-8 -*-
#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

# Seriously python? No RWlock?

from threading import RLock, current_thread
import traceback
import time

import logging
log = logging.getLogger("RWLOCK")

alllocks = []

class RWLock(object):
    def __init__(self, name=""):
        self.name = name
        self.readers = 0
        self.reader_stacks = []
        self.lock = RLock()
        self.reader_lock = RLock()

        self.writer_stacks = []
        self.writer_id = 0

        alllocks.append(self)

    def acquire_read(self, block=True):

        # Hold reader_lock to see if we've already actually got this lock.

        r = self.reader_lock.acquire(block)
        if not r:
            return r

        cti = current_thread().ident
        if cti == self.writer_id or cti in [ x[0] for x in self.reader_stacks ]:
            self.readers += 1
            self.reader_stacks.append((current_thread().ident, traceback.format_stack()))
            self.reader_lock.release()
            return True

        # Release the lock so that if we block on getting the main lock, other
        # threads can still perform the above check and release_read().

        self.reader_lock.release()

        # Get full lock so writers can keep us from getting a lock we don't
        # already hold.

        r = self.lock.acquire(block)
        if not r:
            return r

        # Re-acquire reader_lock so we can manipulate the vars.

        self.reader_lock.acquire()

        self.readers += 1
        self.reader_stacks.append((current_thread().ident, traceback.format_stack()))

        # Release everything.

        self.reader_lock.release()
        self.lock.release()
        return True

    def release_read(self):
        last = False

        self.reader_lock.acquire()
        self.readers -= 1

        for tup in reversed(self.reader_stacks[:]):
            if tup[0] == current_thread().ident:
                self.reader_stacks.remove(tup)
                break

        if self.readers == 0:
            last = True

        self.reader_lock.release()
        return last

    def acquire_write(self, block=True):
        r = self.lock.acquire(block)

        if not r:
            return r

        self.writer_stacks.append(traceback.format_stack())
        self.writer_id = current_thread().ident;

        warned = False

        while self.readers > 0:
            if current_thread().ident in [ x[0] for x in self.reader_stacks ]:
                if not warned:
                    log.debug("WARN: %s holds read, trying to get write on %s", 
                            current_thread().ident, self.name)
                    warned = True

                # Break the deadlock if we're the last reader
                if len(self.reader_stacks) == 1:
                    break

            time.sleep(0.1)
        return True

    def release_write(self):
        last = False

        self.writer_stacks = self.writer_stacks[0:-1]
        if self.writer_stacks == []:
            self.writer_id = 0
            last = True

        self.lock.release()

        return last

def read_lock(lock):
    def _rlock_fn(fn):
        def _rlock(*args, **kwargs):
            lock.acquire_read()
            try:
                return fn(*args, **kwargs)
            finally:
                lock.release_read()
        return _rlock
    return _rlock_fn

def write_lock(lock):
    def _wlock_fn(fn):
        def _wlock(*args, **kwargs):
            lock.acquire_write()
            try:
                return fn(*args, **kwargs)
            finally:
               lock.release_write()
        return _wlock
    return _wlock_fn



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .plugins import PluginHandler, Plugin
from .tag import alltags
from .rwlock import RWLock, read_lock, write_lock
from .locks import feed_lock, tag_lock
from .hooks import call_hook

import traceback
import logging
import json
import time

log = logging.getLogger("FEED")

def dict_id(i):
    if type(i) == dict:
        return i
    return json.loads(i)

class CantoFeeds():
    def __init__(self):
        self.order = []
        self.feeds = {}
        self.dead_feeds = {}

    @write_lock(feed_lock)
    def add_feed(self, URL, feed):
        self.order.append(URL)
        self.feeds[URL] = feed

        # Return old feed object
        if URL in self.dead_feeds:
            del self.dead_feeds[URL]

    @read_lock(feed_lock)
    def get_feed(self, URL):
        if URL in self.feeds:
            return self.feeds[URL]
        if URL in self.dead_feeds:
            return self.dead_feeds[URL]

    @read_lock(feed_lock)
    def get_feeds(self):
        return [ self.get_feed(URL) for URL in self.order]

    @read_lock(feed_lock)
    def items_to_feeds(self, items):
        f = {}
        for i in items:
            d_i = dict_id(i)

            if d_i["URL"] in self.feeds:
                feed = self.feeds[d_i["URL"]]
            else:
                raise Exception("Can't find feed: %s" % d_i["URL"])

            if feed in f:
                f[feed].append(i)
            else:
                f[feed] = [i]
        return f

    def all_parsed(self):
        for URL in self.dead_feeds:
            feed = self.dead_feeds[URL]
            call_hook("daemon_del_tag", [[ "maintag:" + feed.name ]])
            feed.destroy()
        self.dead_feeds = {}

    @write_lock(feed_lock)
    def reset(self):
        self.dead_feeds = self.feeds
        self.feeds = {}
        self.order = []

allfeeds = CantoFeeds()

# Lock helpers

def wlock_all():
    feed_lock.acquire_write()
    for feed in sorted(allfeeds.feeds.keys()):
        allfeeds.feeds[feed].lock.acquire_write()

def wunlock_all():
    for feed in sorted(allfeeds.feeds.keys()):
        allfeeds.feeds[feed].lock.release_write()
    feed_lock.release_write()

def wlock_feeds(fn):
    def _fl(*args):
        wlock_all()
        try:
            return fn(*args)
        finally:
            wunlock_all()
    return _fl

def rlock_all():
    feed_lock.acquire_read()
    for feed in sorted(allfeeds.feeds.keys()):
        allfeeds.feeds[feed].lock.acquire_read()

def runlock_all():
    for feed in sorted(allfeeds.feeds.keys()):
        allfeeds.feeds[feed].lock.release_read()
    feed_lock.release_read()

def rlock_feeds(fn):
    def _fl(*args):
        rlock_all()
        try:
            return fn(*args)
        finally:
            runlock_all()
    return _fl

# feed_objs to enforce
def rlock_feed_objs(objs):
    feed_lock.acquire_read()
    for feed in sorted(allfeeds.feeds.keys()):
        for obj in objs:
            if obj.URL == feed:
                obj.lock.acquire_read()
                break

def runlock_feed_objs(objs):
    for feed in sorted(allfeeds.feeds.keys()):
        for obj in objs:
            if obj.URL == feed:
                obj.lock.release_read()
                break
    feed_lock.release_read()

def stop_feeds():
    for feed in allfeeds.feeds:
        allfeeds.feeds[feed].stopped = True

class DaemonFeedPlugin(Plugin):
    pass

class CantoFeed(PluginHandler):
    def __init__(self, shelf, name, URL, rate, keep_time, keep_unread, **kwargs):
        PluginHandler.__init__(self)

        self.plugin_class = DaemonFeedPlugin
        self.update_plugin_lookups()

        self.shelf = shelf
        self.name = name
        self.URL = URL
        self.rate = rate
        self.keep_time = keep_time
        self.keep_unread = keep_unread
        self.stopped = False

        self.last_update = 0

        # This is held by the update thread, as well as any get / set attribute
        # threads

        self.lock = RWLock()

        self.username = None
        if "username" in kwargs:
            self.username = kwargs["username"]

        self.password = None
        if "password" in kwargs:
            self.password = kwargs["password"]

        allfeeds.add_feed(URL, self)

    def __str__(self):
        return "CantoFeed: %s" % self.name

    # Return { id : { attribute : value .. } .. }

    def get_attributes(self, items, attributes):
        r = {}

        d = self.shelf[self.URL]

        args = [ (dict_id(item)["ID"], item, attributes[item]) for item in items ]
        args.sort()

        got = [ (item["id"], item) for item in d["entries"] ]
        got.sort()

        for item, full_id, needed_attrs in args:
            while got and item > got[0][0]:
                got.pop(0)

            if got and got[0][0] == item:
                attrs = {}
                for a in needed_attrs:
                    if a == "description":
                        real = "summary"
                    else:
                        real = a

                    if real in got[0][1]:
                        attrs[a] = got[0][1][real]
                    else:
                        attrs[a] = ""
                r[full_id] = attrs
                got.pop(0)
            else:
                log.warn("item not found: %s" % item)
                r[full_id] = {}
                for a in needed_attrs:
                    r[full_id][a] = ""
                r[full_id]["title"] = "???"
        return r

    # Given an ID and a dict of attributes, update the disk.
    def set_attributes(self, items, attributes):

        self.lock.acquire_write()

        d = self.shelf[self.URL]

        items_to_remove = []
        tags_to_add = []

        for item in items:
            d_id = dict_id(item)["ID"]

            for d_item in d["entries"]:
                if d_id != d_item["id"]:
                    continue
                for a in attributes[item]:
                    d_item[a] = attributes[item][a]

                items_to_remove.append(d_item)
                tags_to_add += self._tag([d_item])

        self.shelf[self.URL] = d
        self.shelf.update_umod()

        self.lock.release_write()

        self._retag(items_to_remove, tags_to_add, [])

    def _item_id(self, item):
        return json.dumps({ "URL" : self.URL, "ID" : item["id"] })

    def _tag(self, items):
        tags_to_add = []

        for item in items:
            tags_to_add.append((item, "maintag:" + self.name))
            if "canto-tags" in item:
                for user_tag in item["canto-tags"]:
                    log.debug("index adding user tag: %s - %s", user_tag,item["id"])
                    tags_to_add.append((item, user_tag))

        return tags_to_add

    def _retag(self, items_to_remove, tags_to_add, tags_to_remove):
        feed_lock.acquire_read()
        tag_lock.acquire_write()

        for item in items_to_remove:
            alltags.remove_id(self._item_id(item))

        for item, tag in tags_to_add:
            alltags.add_tag(self._item_id(item), tag)

        for item, tag in tags_to_remove:
            alltags.remove_tag(self._item_id(item), tag)

        alltags.do_tag_changes()

        tag_lock.release_write()
        feed_lock.release_read()

    def _keep_olditem(self, olditem):
        ref_time = time.time()

        if "canto_update" not in olditem:
            olditem["canto_update"] = ref_time

        item_time = olditem["canto_update"]

        if "canto-state" in olditem:
            item_state = olditem["canto-state"]
        else:
            item_state = []

        if (ref_time - item_time) < self.keep_time:
            log.debug("Item not over keep_time (%d): %s", 
                    self.keep_time, olditem["id"])
        elif self.keep_unread and "read" not in item_state:
            log.debug("Keeping unread item: %s\n", olditem["id"])
        else:
            log.debug("Discarding: %s", olditem["id"])
            return False
        return True

    # Re-index contents
    # If we have update_contents, use that
    # If not, at least populate self.items from disk.

    # MUST GUARANTEE self.items is in same order as entries on disk.

    def index(self, update_contents):

        # If the daemon is shutting down, discard this update.

        if self.stopped:
            return

        self.lock.acquire_write()

        if self.URL not in self.shelf:
            # Stub empty feed
            log.debug("Previous content not found for %s.", self.URL)
            old_contents = {"entries" : []}
        else:
            old_contents = self.shelf[self.URL]
            log.debug("Fetched previous content for %s.", self.URL)

        new_entries = []

        for i, item in enumerate(update_contents["entries"]):

            # Update canto_update only for freshly seen items.
            item["canto_update"] = update_contents["canto_update"]

            # Attempt to isolate a feed unique ID
            if "id" not in item:
                if "link" in item:
                    item["id"] = item["link"]
                elif "title" in item:
                    item["id"] = item["title"]
                else:
                    log.error("Unable to uniquely ID item: %s" % item)
                    continue

            new_entries.append((i, item["id"], item))

        # Sort by string id
        new_entries.sort(key=lambda x: x[1])

        # Remove duplicates
        last_id = ""
        for x in new_entries[:]:
            if x[1] == last_id:
                new_entries.remove(x)
            else:
                last_id = x[1]

        old_entries = [ (i, item["id"], item) for (i, item) in enumerate(old_contents["entries"])]

        old_entries.sort(key=lambda x: x[1])

        keep_all = new_entries == []

        kept_entries = []

        for x in new_entries:

            # old_entry is really old, see if we should keep or discard

            while old_entries and x[1] > old_entries[0][1]:
                if keep_all or self._keep_olditem(old_entries[0][2]):
                    kept_entries.append(old_entries.pop(0))
                else:
                    old_entries.pop(0)

            # new entry and old entry match, move content over

            if old_entries and x[1] == old_entries[0][1]:
                olditem = old_entries.pop(0)[2]
                for key in olditem:
                    if key == "canto_update":
                        continue
                    elif key.startswith("canto"):
                        x[2][key] = olditem[key]

            # new entry is really new, tell everyone

            else:
                call_hook("daemon_new_item", [self, x[2]])

        # Resort lists by place, instead of string id
        new_entries.sort()
        old_entries.sort()

        if keep_all:
            kept_entries += old_entries
        else:
            for x in old_entries:
                if self._keep_olditem(x[2]):
                    kept_entries.append(x)

        kept_entries.sort()
        new_entries += kept_entries

        update_contents["entries"] = [ x[2] for x in new_entries ]

        tags_to_add = self._tag(update_contents["entries"])
        tags_to_remove = []
        remove_items = []

        # Allow plugins to add items prior to running the editing functions
        # so that the editing functions are guaranteed the full list.

        for attr in list(self.plugin_attrs.keys()):
            if not attr.startswith("additems_"):
                continue

            try:
                a = getattr(self, attr)
                tags_to_add, tags_to_remove, remove_items = a(self, update_contents, tags_to_add, tags_to_remove, remove_items)
            except:
                log.error("Error running feed item adding plugin")
                log.error(traceback.format_exc())

        # Allow plugins DaemonFeedPlugins defining edit_* functions to have a
        # crack at the contents before we commit to disk.

        for attr in list(self.plugin_attrs.keys()):
            if not attr.startswith("edit_"):
                continue

            try:
                a = getattr(self, attr)
                tags_to_add, tags_to_remove, remove_items = a(self, update_contents, tags_to_add, tags_to_remove, remove_items)
            except:
                log.error("Error running feed editing plugin")
                log.error(traceback.format_exc())

        if not self.stopped:
            # Commit the updates to disk.

            self.shelf[self.URL] = update_contents

            self.lock.release_write()

            self._retag(old_contents["entries"] + remove_items, tags_to_add, tags_to_remove)
        else:
            self.lock.release_write()

    def destroy(self):
        # Check for existence in case of delete quickly
        # after add.

        self.stopped = True
        if self.URL in self.shelf:
            del self.shelf[self.URL]



# Canto rsync Plugin
# by Jack Miller
# v1.1

# This implements a lightweight remote sync based around rsync to a remote
# server, or copying to mounted filesystem, etc.

ENABLED = False
#ENABLED = True

# SSH
# For ssh based rsync (remote hosts) you should have key authentication setup
# so it runs without prompting for a password.

#SYNC_LOCATION = "user@host:"

# Dropbox, assuming you have dropbox running
#SYNC_LOCATION = "~/Dropbox/"

# Mount / NFS / sshfs etc.
#SYNC_LOCATION = "/mnt/wherever/"

# Synchronization interval in seconds
INTERVAL = 5 * 60

# How long, in seconds, we should wait for the initial sync. Setting to 0 will
# cause a sync to occur before any other items can be read from disk, which
# ensures you won't see any old items, but also means a full sync has to occur
# before any items make it to the client and causes a long delay on startup.

INITIAL_SYNC = 30

#============================================
# Probably won't need to change these.

# rsync
#  -a (archive mode) to preserve times / perms
#  -v (verbose) to output interesting log info
#  -z (compress) to save bandwidth

CMD = [ "rsync", "-avz"]

targets = { "db" : ".cantofeeds",
            "conf" : ".cantoconf"
}

from canto_next.plugins import check_program

check_program("canto-daemon", "canto-remote")

if not ENABLED:
    raise Exception("Plugin disabled.")

from canto_next.hooks import on_hook, call_hook
from canto_next.canto_backend import DaemonBackendPlugin
from canto_next.remote import DaemonRemotePlugin

from canto_next.config import parse_locks, parse_unlocks, config
from canto_next.locks import config_lock, feed_lock
from canto_next.feed import wlock_all, wunlock_all, rlock_all, runlock_all, allfeeds
from canto_next.tag import alltags

from tempfile import mkstemp
import subprocess
import logging
import shutil
import gzip
import json
import time
import os

log = logging.getLogger("SYNC-RSYNC")

class CantoFileSync(DaemonBackendPlugin):
    def __init__(self, backend):
        self.plugin_attrs = {
                "cmd_sync" : self.cmd_sync,
                "cmd_syncto" : self.cmd_syncto
        }

        self.backend = backend

        # Plugin __init__ happens extremely early so that plugin types can be
        # used in validating configuration, etc. We use the daemon_serving hook
        # to do our work after the config and storage is setup.

        on_hook("daemon_serving", self.setup)

    def setup(self):

        # Use setattributes and setconfigs commands to determine that we are the fresh
        # copy that should be synchronized.

        on_hook("daemon_end_loop", self.loop)
        on_hook("daemon_pre_setconfigs", self.pre_setconfigs)
        on_hook("daemon_pre_setattributes", self.pre_setattributes)
        on_hook("daemon_exit", self.cmd_syncto)

        self.reset()

        # sync will grab files, check the timediff on the file if the file is
        # actually newer (like we failed to sync last time) then it will set
        # fresh_config and do a syncto.

        self.sync_ts = 0
        if (INITIAL_SYNC == 0):
            self.cmd_sync()
        elif (INITIAL_SYNC < INTERVAL):
            self.sync_ts = time.time() - (INTERVAL - INITIAL_SYNC)

    def reset(self):
        self.fresh_config = False
        self.sent_config = False

        self.fresh_content = False
        self.sent_content = False

    # Use hooks to determine when we need to copy stuff.

    def pre_setattributes(self, socket, args):
        self.fresh_content = True

    def pre_setconfigs(self, socket, args):
        self.fresh_config = True

    # Open a shelf at path, determine if it's been changed more recently than
    # our current shelf.

    def time_diff(self, path):
        log.debug("Checking if %s is older than our shelf.", path)

        try:
            fp = gzip.open(path, "rt", 9, "UTF-8")
            s = json.load(fp)
            fp.close()
        except:
            # If something messed up, assume that the sync failed and
            # pretend that we're newer anyway.
            return -1

        if "control" in s and "canto-user-modified" in s["control"]:
            remote_stamp = s["control"]["canto-user-modified"]
        else:
            log.debug("Remote has no timestamp")
            return -1

        rlock_all()
        if "control" in self.backend.shelf and "canto-user-modified" in self.backend.shelf["control"]:
            local_stamp = self.backend.shelf["control"]["canto-user-modified"]
            runlock_all()
        else:
            log.debug("We have no timestamp")
            runlock_all()
            return 1

        if remote_stamp > local_stamp:
            log.debug("db: We are older")
        elif remote_stamp == local_stamp:
            log.debug("db: We are equal")
        else:
            log.debug("db: We are newer")

        return remote_stamp - local_stamp

    def cmd_syncto(self, socket = None, args = None):
        if self.fresh_content:
            f, fname = mkstemp()
            os.close(f)

            # Lock feeds to make sure nothing's in flight
            wlock_all()

            # Sync the shelf so it's all on disk

            self.backend.shelf.sync()

            shutil.copyfile(self.backend.feed_path, fname)

            # Let everything else continue
            wunlock_all()

            call_hook("daemon_syncto", [ "db", fname ])

            # Cleanup temp file
            os.unlink(fname)

            self.fresh_content = False
            self.sent_content = True

        if self.fresh_config:
            f, fname = mkstemp()
            os.close(f)

            config_lock.acquire_read()
            shutil.copyfile(self.backend.conf_path, fname)
            config_lock.release_read()

            call_hook("daemon_syncto", [ "conf", fname ])

            os.unlink(fname)

            self.fresh_config = False
            self.sent_config = True

    def cmd_sync(self, socket = None, args = None):
        needs_syncto = False

        if not self.sent_config:
            f, fname = mkstemp()
            os.close(f)

            call_hook("daemon_syncfrom", [ "conf", fname ])

            conf_stat = os.stat(self.backend.conf_path)
            sync_stat = os.stat(fname)

            log.debug('conf: %s sync: %s' % (conf_stat.st_mtime, sync_stat.st_mtime))

            diff = sync_stat.st_mtime - conf_stat.st_mtime

            # Will be empty tempfile if syncfrom failed.

            if sync_stat.st_size != 0:
                if diff > 0:
                    log.debug("conf: We are older")
                    parse_locks()
                    shutil.move(fname, self.backend.conf_path)
                    config.parse()
                    parse_unlocks()

                    # Echo these changes to all connected sockets that care
                    for socket in self.backend.watches["config"]:
                        self.backend.in_configs({}, socket)

                elif diff == 0:
                    log.debug("conf: We are equal")
                    os.unlink(fname)
                else:
                    log.debug("conf: We are newer")
                    os.unlink(fname)
                    self.fresh_config = True
                    needs_syncto = True
            else:
                os.unlink(fname)

        if not self.sent_content:
            f, fname = mkstemp()
            os.close(f)

            call_hook("daemon_syncfrom", [ "db", fname ])

            diff = self.time_diff(fname)

            if diff > 0:
                # Lock feeds to make sure nothing's in flight
                wlock_all()

                # Close the file so we can replace it.
                self.backend.shelf.close()

                shutil.move(fname, self.backend.feed_path)

                self.backend.shelf.open()

                # Clear out all of the currently tagged items. Usually on
                # update, we're able to discard items that we have in old
                # content, but aren't in new. But since we just replaced all of
                # our old content with a totally fresh copy, we might not know
                # they exist. Can't use reset() because we don't want to lose
                # configuration.

                alltags.clear_tags()

                # First half of wunlock_all, release these locks so
                # fetch threads can get locks

                for feed in sorted(allfeeds.feeds.keys()):
                    allfeeds.feeds[feed].lock.release_write()

                # Complete wunlock_all()
                feed_lock.release_write()

                # Force feeds to be repopulated from disk, which will handle
                # communicating changes to connections

                self.backend.fetch.fetch(True, True)
                self.backend.fetch.reap(True)

            # Equal, just clear it up

            elif diff == 0:
                os.unlink(fname)

            # If we're actually newer on a syncfrom then make syncto happen
            # next time. This can happen on init.

            else:
                os.unlink(fname)
                self.fresh_content = True
                needs_syncto = True

        if needs_syncto:
            self.cmd_syncto()

        self.reset()

    def loop(self):
        ts = time.time()
        if (ts - self.sync_ts >= INTERVAL):
            self.cmd_sync()
            self.sync_ts = ts

class RemoteSync(DaemonRemotePlugin):
    def __init__(self, remote):
        self.plugin_attrs = { "cmd_sync" : self.cmd_sync }
        self.remote = remote

        on_hook("remote_print_commands", self.print_sync_commands)

    def print_sync_commands(self):
        print("\nSync Plugin")
        print("\tsync - sync the daemon")

    def cmd_sync(self):
        """USAGE: canto-remote sync
    Synchronize this daemon with a remote daemon"""
        self.remote.write("SYNC", {})

# Each of these gets called with a "target" (i.e. a type of file we want to
# sync) and a temporary filename to either copy to somewhere else or overwrite.

# NOTE: The logic for whether this file actually gets used is in sync.py. For
# the feeds database, it takes the last user modification into account because
# any db with a running daemon is going to be modified often by new feed info,
# making mtime worthless. For the config, however, it's only written when a
# change has been made, so mtime should be sufficient. This is why we use -a to
# rsync.

def rsync_to(target, fname):

    if target in targets:
        cmd = CMD + [ fname, SYNC_LOCATION + targets[target] ]
    else:
        log.warn("Unknown file to sync: %s" % target)
        return

    log.debug("Syncto cmd: %s", cmd)
    
    try:
        out = subprocess.check_output(cmd)
    except Exception as e:
        log.warn("Command %s : %s" % (cmd, e))
    else:
        log.debug("Syncto output: %s", out)

def rsync_from(target, fname):
    if target in targets:
        cmd = CMD + [ SYNC_LOCATION + targets[target], fname ]
    else:
        log.warn("Unknown file to sync: %s" % target)
        return

    log.debug("Syncfrom cmd: %s", cmd)

    try:
        out = subprocess.check_output(cmd)
    except Exception as e:
        log.warn("Command %s : %s" % (cmd, e))
    else:
        log.debug("Syncfrom output: %s", out)

on_hook("daemon_syncfrom", rsync_from)
on_hook("daemon_syncto", rsync_to)



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import traceback
import logging

log = logging.getLogger("HOOKS")

hooks = {}

def on_hook(hook, func, key=None):
    if key != None and type(key) != str:
        key = str(key)

    if hook in hooks:
        if key in hooks[hook]:
            hooks[hook][key].append(func)
        else:
            hooks[hook][key] = [ func ]
    else:
        hooks[hook] = { key : [ func ] }

def remove_hook(hook, func):
    for key in list(hooks[hook].keys()):
        while func in hooks[hook][key]:
            hooks[hook][key].remove(func)
        if hooks[hook][key] == []:
            del hooks[hook][key]
            if hooks[hook] == {}:
                del hooks[hook]

def unhook_all(key):
    if key != None and type(key) != str:
        key = str(key)
    for hook in list(hooks.keys()):
        if key in hooks[hook]:
            del hooks[hook][key]
            if hooks[hook] == {}:
                del hooks[hook]

def call_hook(hook, args):
    if hook in hooks:
        for key in list(hooks[hook].keys()):
            try:
                for func in hooks[hook][key][:]:
                    try:
                        func(*args)
                    except:
                        log.error("Error calling hook %s (func: %s args: %s)" % (hook, func, args))
                        log.error(traceback.format_exc())
            except:
                pass



# Canto Inoreader Plugin
# by Jack Miller
# v0.4

# DEPENDENCIES

# This plugin requires the 'requests' module, which can usually be found in
# your distro's package manager as python3-requests (or python-requests on
# Arch).

# IMPORTANT NOTES

# - When this plugin is enabled, canto will synchronize your subscribed feeds
# with Inoreader. If you've been using canto, you should export your feeds
# (canto-remote export > feeds.opml) and import them into Inoreader if you
# don't want to lose your feeds because Inoreader's info is assumed to be more
# correct than ours.
#
# - Feed subscriptions are only synchronized *from Inoreader* on startup, so if
# you add / remove feeds with Inoreader, you will have to restart the daemon to
# pickup the changes. Adding or removing feeds with canto works properly all
# the time.
#
# - You must have a standard Inoreader account, not an OAuth (Google/Facebook
# login).

# CONFIGURATION

# Inoreader credentials

EMAIL="somebody@somewhere.com"
PASSWORD="passw0rd"

# You don't *have* to change these, but the API is rate limited. So if you want
# to avoid rate limit issues, register your own application Preferences ->
# Developer options on the Inoreader site and replace these.

APP_ID="1000001299"
APP_KEY="i0UOUtLQjj2WTre8WA3a9GWt_cgDhpkO"

BASE_URL="https://www.inoreader.com/reader/"

# === You shouldn't have to change anything past this line. ===

from canto_next.plugins import check_program

check_program("canto-daemon")

from canto_next.fetch import DaemonFetchThreadPlugin
from canto_next.feed import DaemonFeedPlugin, allfeeds
from canto_next.hooks import call_hook, on_hook
from canto_next.config import config

from urllib.parse import urlencode, quote
import traceback
import requests
import logging
import time
import json

log = logging.getLogger("SYNC-INOREADER")

class InoreaderReqFailed(Exception):
    pass

class InoreaderAuthFailed(Exception):
    pass

class CantoInoreaderAPI():
    def __init__(self):
        self.extra_headers = {
                "User-Agent" : "Canto/0.9.0 + http://codezen.org/canto-ng",
                "AppKey" : APP_KEY,
                "AppID" : APP_ID,
        }

        try:
            self.authorization = self.auth()
        except:
            self.authorization = None

        self.dead = False

        self.add_tags_queued = {}
        self.del_tags_queued = {}

    def auth(self):
        headers = self.extra_headers.copy()
        headers['Email'] = EMAIL
        headers['Passwd'] = PASSWORD

        try:
            r = requests.get("https://www.inoreader.com/accounts/ClientLogin", headers, timeout=1)
        except Exception as e:
            raise InoreaderReqFailed(str(e))

        if r.status_code != 200:
            raise InoreaderAuthFailed("Failed to authorize: [%s] %s" % (r.status_code, r.text))

        for line in r.text.splitlines():
            if line.startswith("Auth="):
                log.debug("authorization: %s", line[5:])
                return line[5:]

        raise InoreaderAuthFailed("Failed to find Auth= in auth response")

    def inoreader_req(self, path, query = {}):
        tries = 3
        r = {}

        while tries and not self.dead:
            tries -= 1
            if not self.authorization:
                try:
                    self.authorization = self.auth()
                except InoreaderReqFailed as e:
                    log.debug("Auth request failed: %s", e)
                    continue
                except InoreaderAuthFailed:
                    log.error("Inoreader authorization failed, please check your credentials in sync-inoreader.py")
                    self.dead = True
                    raise

            headers = self.extra_headers.copy()
            headers["Authorization"] = "GoogleLogin auth=" + self.authorization

            try:
                r = requests.get(BASE_URL + path, params=query, headers=headers, timeout=1)
            except requests.exceptions.Timeout:
                raise InoreaderReqFailed

            if r.status_code != 200:
                log.debug("STATUS %s", r.status_code)
                log.debug(r.headers)
                log.debug(r.text)
            else:
                return r

            # No authorization, attempt to get another code on the next try.

            if r.status_code == 401:
                self.authorization = None
            elif r.status_code == 429:
                log.error("Inoreader rate limit reached.")
                self.dead = True
            elif r.status_code == 503:
                log.error("Inoreader appears down, state may be lost")

        raise InoreaderReqFailed

    # Convert special tags into /state/com.google/tag and others into
    # /label/tag, useful when matching without knowing the user.

    def full_ino_tag_suffix(self, tag):
        if tag in ["read", "starred", "fresh"]:
            return "/state/com.google/" + tag
        return "/label/" + tag

    # Add the user/- prefix to go upstream to Inoreader.

    def full_ino_tag(self, tag):
        return "user/-" + self.full_ino_tag_suffix(tag)

    # Do the opposite, convert an Inoreader tag into a natural name.  (i.e.)
    # /user/whatever/state/com.google/read -> read

    def strip_ino_tag(self, tag):
        tag = tag.split("/", 3)
        if tag[2] == "state":
            return tag[3].split("/", 1)[1]
        return tag[3]

    # Return whether Inoreader data includes this natural tag

    def has_tag(self, item, tag):
        if "canto_inoreader_categories" not in item:
            return False

        suff = self.full_ino_tag_suffix(tag)
        for category in item["canto_inoreader_categories"]:
            if category.endswith(suff):
                return True
        return False

    def add_tag(self, item, tag):
        ino_id = item["canto_inoreader_id"]
        if not self.has_tag(item, tag):
            if tag in self.add_tags_queued:
                self.add_tags_queued[tag].append(ino_id)
            else:
                self.add_tags_queued[tag] = [ino_id]

    def remove_tag(self, item, tag):
        ino_id = item["canto_inoreader_id"]
        if self.has_tag(item, tag):
            if tag in self.del_tags_queued:
                self.del_tags_queued[tag].append(ino_id)
            else:
                self.del_tags_queued[tag] = [ino_id]

    def _urllimit(self, prefix, ino_ids):
        t = prefix
        l = len(t)

        for i, ino_id in enumerate(ino_ids):
            if l + len(ino_id) > 2048:
                self.inoreader_req(t)
                return ino_ids[i:]
            t += ino_id
            l += len(ino_id)

        self.inoreader_req(t)
        return []

    def flush_changes(self):
        for key in self.add_tags_queued:
            to_add = [ "&i=" + quote(x) for x in self.add_tags_queued[key]]
            while to_add:
                to_add = self._urllimit("api/0/edit-tag?a=" + quote(self.full_ino_tag(key)), to_add)

        for key in self.del_tags_queued:
            to_del = [ "&i=" + quote(x) for x in self.del_tags_queued[key]]
            while to_del:
                to_del = self._urllimit("api/0/edit-tag?r=" + quote(self.full_ino_tag(key)), to_del)

        self.add_tags_queued = {}
        self.del_tags_queued = {}

    def get_subs(self):
        return self.inoreader_req("api/0/subscription/list").json()["subscriptions"]

    def add_sub(self, feed_url, title):
        query = {
            "ac" : "subscribe",
            "s" : "feed/" + feed_url,
            "t" : title
        }

        self.inoreader_req("api/0/subscription/edit", query)

    def del_sub(self, feed_url):
        query = {
            "ac" : "unsubscribe",
            "s" : "feed/" + feed_url
        }

        self.inoreader_req("api/0/subscription/edit", query)

api = CantoInoreaderAPI()

# Given a change set, and the current attributes of a canto item, tell
# Inoreader about it.

def sync_state_to(changes, attrs, add_only = False):
    if "canto-state" in changes:
        if "read" in changes["canto-state"]:
            api.add_tag(attrs, "read")
        elif not add_only:
            if api.has_tag(attrs, "read"):
                api.remove_tag(attrs, "read")

    if "canto-tags" in changes:
        for tag in changes["canto-tags"]:
            tag = tag.split(":", 1)[1] # strip user: or category: prefix
            if not api.has_tag(attrs, tag):
                api.add_tag(attrs, tag)

        if add_only:
            return

        for tag in attrs["canto_inoreader_categories"]:
            tag = api.strip_ino_tag(tag)
            if "user:" + tag not in changes["canto-tags"]:
                api.remove_tag(attrs, tag)

class CantoFeedInoReader(DaemonFeedPlugin):
    def __init__(self, feed):
        self.plugin_attrs = { "edit_inoreader_sync" : self.edit_inoreader_sync,
                "additems_inoreader" : self.additems_inoreader }
        self.feed = feed
        self.ino_data = None

    def _list_add(self, item, attr, new):
        if attr not in item:
            item[attr] = [ new ]
        elif new not in item[attr]:
            item[attr].append(new)

    def add_utag(self, item, tags_to_add, tag):
        self._list_add(item, "canto-tags", "user:" + tag)
        tags_to_add.append((item, "user:" + tag))

    def add_state(self, item, state):
        self._list_add(item, "canto-state", state)

    def additems_inoreader(self, feed, newcontent, tags_to_add, tags_to_remove, remove_items):
        stream_id = quote("feed/" + feed.URL, [])

        query = { "n" : 1000 }

        # Collect all of the items

        self.ino_data = []

        content_path = "api/0/stream/contents/" + stream_id

        try:
            r = api.inoreader_req(content_path, query).json()
            self.ino_data.extend(r["items"])
        except (InoreaderAuthFailed, InoreaderReqFailed):
            return (tags_to_add, tags_to_remove, remove_items)
        except Exception as e:
            log.debug("EXCEPT: %s", traceback.format_exc())
            raise

        # Find items that were inserted last time, and remove them, potentially
        # adding them to our fresh Inoreader data.

        # This keeps us from getting dupes when Inoreader finds an item, we
        # insert it, and then a real copy comes to canto but canto doesn't
        # detect the dupe since the ids are different.

        for canto_entry in newcontent["entries"][:]:
            if "canto-from-inoreader" not in canto_entry:
                continue

            remove_items.append(canto_entry)
            tags_to_add = [ x for x in tags_to_add if x[0] != canto_entry]

            newcontent["entries"].remove(canto_entry)

            for ino_entry in self.ino_data[:]:
                if canto_entry["id"] == ino_entry["id"]:
                    break
            else:
                self.ino_data.append(canto_entry)

        # Now insert (or re-insert) items that aren't already in our data.

        # NOTE: It's okay if re-inserted items are also in remove_ids, since
        # that's processed first, and will be cancelled out by adding the tags
        # afterwards.

        for ino_entry in self.ino_data:
            for canto_entry in newcontent["entries"]:
                if ino_entry["canonical"][0]["href"] != canto_entry["link"]:
                    continue
                if ino_entry["id"] == canto_entry["id"]:
                    canto_entry["canto-from-inoreader"] = True
                break
            else:
                if "canto-from-inoreader" not in ino_entry:
                    # feedparser compatibility
                    ino_entry["summary"] = ino_entry["summary"]["content"]
                    ino_entry["link"] = ino_entry["canonical"][0]["href"]

                    # mark this item as from inoreader (missing from feed)
                    ino_entry["canto-from-inoreader"] = True

                newcontent["entries"].append(ino_entry)
                tags_to_add.append((ino_entry, "maintag:" + feed.name ))

        return (tags_to_add, tags_to_remove, remove_items)

    def edit_inoreader_sync(self, feed, newcontent, tags_to_add, tags_to_remove, remove_items):

        # Add inoreader_id/categories information to the items

        # This is very similar to the loop in additems_inoreader, but needs to
        # be separate in case other plugins add items that inoreader might
        # track.

        for ino_entry in self.ino_data:
            for canto_entry in newcontent["entries"][:]:
                if ino_entry["canonical"][0]["href"] != canto_entry["link"]:
                    continue
                canto_entry["canto_inoreader_id"] = ino_entry["id"]
                canto_entry["canto_inoreader_categories"] = ino_entry["categories"]
                break

        # Drop the data.
        self.ino_data = None

        for entry in newcontent["entries"]:
            # If we didn't get an id for this item, skip it

            if "canto_inoreader_id" not in entry:
                continue

            for category in entry["canto_inoreader_categories"]:
                if category.endswith("/state/com.google/read"):
                    self.add_state(entry, "read")
                    continue

                cat = category.split("/", 3)
                if len(cat) < 4:
                    log.debug("Weird category? %s", cat)
                    continue

                if cat[2] == "state":
                    if cat[3] == "com.google/starred":
                        self.add_utag(entry, tags_to_add, "starred")
                elif cat[2] == "label":
                    self.add_utag(entry, tags_to_add, cat[3])

            # If this is the first time we've paired an item up with its
            # Inoreader data, our state is better, so sync it to Inoreader, and
            # then skip the remainder of the logic to remove canto state/tags

            if "canto-inoreader-sync" not in entry:
                sync_state_to(entry, entry, True)
                entry["canto-inoreader-sync"] = True
                continue

            if "canto-state" not in entry or type(entry["canto-state"]) != list:
                continue

            # It appears that if an item is "fresh" it will resist all attempts
            # to set it as read?

            if "read" in entry["canto-state"] and not\
                    (api.has_tag(entry, "read") or api.has_tag(entry, "fresh")):
                log.debug("Marking unread from Inoreader")
                entry["canto-state"].remove("read")

            if "canto-tags" not in entry or type(entry["canto-tags"]) != list:
                continue

            for tag in entry["canto-tags"][:]:
                if not api.has_tag(entry, tag.split(":", 1)[1]):
                    entry["canto-tags"].remove(tag)
                    tags_to_remove.append((entry, tag))

        api.flush_changes()
        return (tags_to_add, tags_to_remove, remove_items)

# For canto communicating to Inoreader, we tap into the relevant hooks to
# pickup state / tag changes, and convert that into Inoreader API calls.

def post_setattributes(socket, args):
    for item_id in args.keys():
        dict_id = json.loads(item_id)

        feed = allfeeds.get_feed(dict_id["URL"])

        attrs = feed.get_attributes([item_id], { item_id :\
                ["canto_inoreader_id", "canto_inoreader_categories", "canto-state", "canto-tags"] })
        attrs = attrs[item_id]

        # If the canto_inoreader_id isn't right (likely empty since get_attributes
        # will sub in "") then skip synchronizing this item.

        ino_id = attrs["canto_inoreader_id"]
        if not ino_id.startswith("tag:google.com,2005:reader/item/"):
            continue

        sync_state_to(args[item_id], attrs)

    api.flush_changes()

on_hook("daemon_post_setattributes", post_setattributes)

def post_setconfigs(socket, args):
    if "feeds" in args:
        for feed in args["feeds"]:
            api.add_sub(feed["url"], feed["name"])

on_hook("daemon_post_setconfigs", post_setconfigs)

def post_delconfigs(socket, args):
    if "feeds" in args:
        for feed in args["feeds"]:
            api.del_sub(feed["url"])

on_hook("daemon_post_delconfigs", post_delconfigs)

# Do the initial feed synchronization. This only occurs once per run, and
# assumes Inoreader knows everything.

def on_daemon_serving():
    log.debug("Synchronizing subscriptions.")
    ino_subs = api.get_subs()

    for c_feed in config.json["feeds"]:
        url = c_feed["url"]

        for sub in ino_subs:
            if sub["url"] == url:
                break
        else:
            log.debug("Old feed: %s", url)
            call_hook("daemon_del_configs", [ None, { "feeds" : [ c_feed ] } ] )

    for sub in ino_subs:
        url = sub["url"]
        name = sub["title"]

        for c_feed in config.json["feeds"]:
            if c_feed["url"] == url:
                break
            if c_feed["name"] == name:
                log.info("Found feed with same name, but not URL? Skipping.")
                break
        else:
            log.debug("New feed: %s", url)
            call_hook("daemon_set_configs", [ None, { "feeds" : [ { "name" : name, "url" : url } ] } ])


on_hook("daemon_serving", on_daemon_serving)



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .locks import feed_lock, config_lock, tag_lock, watch_lock
from .encoding import locale_enc
from .transform import eval_transform
from .feed import allfeeds, CantoFeed
from .tag import alltags

import traceback
import logging
import codecs
import json
import os
import re

log = logging.getLogger("CONFIG")

default_config =\
{
        "defaults" :
        {
            "rate" : 10,
            "keep_time" : 86400,
            "keep_unread" : False,
            "global_transform" : "filter_read"
        },

        "feeds" : [
            {
                "name" : "Canto",
                "url" : "http://codezen.org/canto-ng/feed/"
            },
            {
                "name" : "Slashdot",
                "url" : "http://rss.slashdot.org/slashdot/Slashdot"
            },
            {
                "name" : "Reddit",
                "url": "http://reddit.com/.rss"
            }
        ]
}

def parse_locks():
    config_lock.acquire_write()
    feed_lock.acquire_write()
    tag_lock.acquire_write()
    watch_lock.acquire_read()

def parse_unlocks():
    config_lock.release_write()
    feed_lock.release_write()
    tag_lock.release_write()
    watch_lock.release_read()

class CantoConfig():
    def init(self, filename, shelf):
        self.filename = filename
        self.shelf = shelf
        self.json = {}

        self.defaults_validators = [
                ("rate", self.validate_int, False),
                ("keep_time", self.validate_int, False),
                ("keep_unread", self.validate_bool, False),
                ("global_transform", self.validate_set_transform, False),
        ]

        self.defaults_defaults = {
                "rate" : 10,
                "keep_time" : 86400,
                "keep_unread" : False,
                "global_transform" : "None",
        }

        self.feed_validators = [
                ("name", self.validate_unique_feed_name, True),
                ("url", self.validate_unique_url, True),
                ("rate", self.validate_int, False),
                ("keep_time", self.validate_int, False),
                ("keep_unread", self.validate_bool, False),
                ("username", self.validate_string, False),
                ("password", self.validate_string, False),
        ]

        self.feed_defaults = {}

        self.tag_validators = [
                ("transform", self.validate_set_transform, False),
                ("extra_tags", self.validate_string_list, False),
        ]

        self.tag_defaults = {}

    def reset(self):
        allfeeds.reset()
        alltags.reset()

        self.errors = {}
        self.final = {}

        # Accumulators for verifying uniqueness
        self.urls = []
        self.feed_names = []

    def parse(self, fromfile=True, changes={}):
        parse_locks()

        # Since we host client config too, check if
        # we should do a reparse.

        we_care = False
        for header in [ "feeds", "tags", "defaults" ]:
            if header in changes:
                if header == "tags":
                    for tag in changes["tags"]:
                        if list(changes["tags"][tag].keys()) != [ "collapsed" ]:
                            we_care = True
                    if we_care:
                        break
                else:
                    we_care = True
                    break

        if fromfile or we_care:
            self.reset()
            if fromfile:
                self.read_config()
            if self.validate():
                self.instantiate()
                if not fromfile:
                    self.write()
        elif not fromfile:
            self.write()

        parse_unlocks()

    def read_config(self):
        if not os.path.exists(self.filename):
            log.info("No config found, writing default.")
            self.json = default_config.copy()
            self.write()

        c = codecs.open(self.filename, "rb", locale_enc)
        self.json = json.load(c)
        c.close()

        log.info("Read %s" % self.filename)
        log.debug("Parsed into: %s", self.json)

    def error(self, ident, val, error):
        if ident in self.errors:
            self.errors[ident].append((val, error))
        else:
            self.errors[ident] = [(val, error)]

    def _validate_unique(self, ident, value, accumulator, desc):
        if not self.validate_string(ident, value):
            return False

        if value in accumulator:
            self.error(ident, value, "%s already used!" % (desc,))
            return False

        accumulator.append(value)
        return (True, value)

    def validate_unique_url(self, ident, value):
        return self._validate_unique(ident, value, self.urls, "URL")

    def validate_unique_feed_name(self, ident, value):
        return self._validate_unique(ident, value, self.feed_names, "Feed name")

    def validate_bool(self, ident, value):
        if type(value) != bool:
            self.error(ident, value, "Not boolean!")
            return False
        return (True, value)

    def validate_int(self, ident, value):
        if type(value) != int:
            self.error(ident, value, "Not integer!")
            return False
        return (True, value)

    def validate_string(self, ident, value):
        if type(value) != str:
            self.error(ident, value, "Not unicode!")
            return False
        return (True, value)

    def validate_string_list(self, ident, value):
        if type(value) != list:
            self.error(ident, value, "Not list!")
            return False

        for idx, item in enumerate(value):
            item_ident = ident + ("[%d]" % idx)
            if not self.validate_string(item_ident, item):
                return False

        return (True, value)

    # Unfortunately, this must return the value, so that the JSON doesn't get
    # tainted with non-serializable values.

    def validate_set_transform(self, ident, value):
        try:
            r = eval_transform(value)
        except Exception as e:
            tb = traceback.format_exc()
            msg = "\n" + "".join(tb)
            self.error(ident, value, "Invalid transform" + msg)
            return (True, "None")

        return (True, value)

    def validate_dict(self, ident_prefix, d, validators):
        section_invalidated = False
        for rgx, validator, required in validators:
            r = re.compile(rgx)

            found = False

            for opt in list(d.keys()):
                match = r.match(opt)
                if not match:
                    continue

                found = True
                ident = ident_prefix + ("[%s]" % opt)

                ret = validator(ident, d[opt])
                if not ret:
                    if required:
                        self.error(ident, d[opt],\
                                "Set but invalid and required!")
                        section_invalidated = True
                    else:
                        self.error(ident, d[opt], "Set but invalid!")
                        del d[opt]
                else:
                    # NOTE: we're ignoring the first tuple, it should
                    # always be True. If it wasn't for the fact that (val,)
                    # looks terrible that could also be returned from the
                    # validators.

                    d[opt] = ret[1]

            if not found and required:
                ident = ident_prefix + "[%s]" % rgx
                self.error(ident, None,\
                        "No matching value found on required option!")
                section_invalidated = True

            if section_invalidated:
                break

        return not section_invalidated

    # Validate validates only what exists in self.final, it does not make
    # substitutions for defaults. That's done on instantiation.

    def validate(self):
        # Because we have to ensure that all items in the JSON are
        # simple, we can do this cheap deepcopy intead of importing
        # copy or doing it ourselves.

        self.final = eval(repr(self.json), {}, {})

        if "defaults" in self.final:
            good = self.validate_dict("[defaults]", self.final["defaults"],
                    self.defaults_validators)
            if not good:
                del self.final["defaults"]

        if "tags" in self.final and not self.errors:
            for tag in list(self.final["tags"].keys()):
                good = self.validate_dict("[tags][" + tag + "]", self.final["tags"][tag],
                        self.tag_validators)
                if not good:
                    del self.final["tags"][tag]

        if "feeds" in self.final and not self.errors:
            for i, feed in enumerate(self.final["feeds"][:]):
                good = self.validate_dict("[feeds][%s]" % i, feed,
                        self.feed_validators)
                if not good:
                    self.final["feeds"].remove(feed)

        if self.errors:
            log.error("ERRORS:")
            for key in list(self.errors.keys()):
                log.error("%s:" % key)
                for value, error in self.errors[key]:
                    log.error("\t%s -> %s" % (value, error))
            return False

        log.info("Validated: %s" % self.final)
        return True

    # Create Tag and Feed objects based on final validated config

    def instantiate(self):

        if "defaults" in self.final:
            for k in self.defaults_defaults.keys():
                if k not in self.final["defaults"]:
                    self.final["defaults"][k] = self.defaults_defaults[k]
        else:
            self.final["defaults"] = self.defaults_defaults.copy()

        if "tags" in self.final:
            for tag in self.final["tags"]:
                defs = self.final["tags"][tag]

                if "transform" not in defs:
                    defs["transform"] = "None"

                defs["transform"] = eval_transform(defs["transform"])

                if "extra_tags" not in defs:
                    defs["extra_tags"] = []

                alltags.tag_transform(tag, defs["transform"])
                alltags.set_extra_tags(tag, defs["extra_tags"])

        # Feeds must be instantiated *after* tags, so tag settings like extra_tags
        # can rely on getting an add_tag for each item after all tag settings have
        # been handled.

        if "feeds" in self.final:
            for feed in self.final["feeds"]:

                # Mandatory arguments to CantoFeed
                for k in [ "rate", "keep_time", "keep_unread" ]:
                    if k not in feed:
                        feed[k] = self.final["defaults"][k]

                # Optional arguments in kwargs
                kws = {}
                for k in ["password", "username"]:
                    if k in feed:
                        kws[k] = feed[k]

                feed = CantoFeed(self.shelf, feed["name"],\
                        feed["url"], feed["rate"], feed["keep_time"], feed["keep_unread"], **kws)

        # Set global transform.

        self.global_transform = eval_transform(\
                self.final["defaults"]["global_transform"])

    # Delete settings from the JSON. Any key equal to "DELETE" will be removed,
    # keys that are lists will items removed if specified.

    def _delete(self, deletions, current):
        for key in list(deletions.keys()):

            # Nothing to do.

            if key not in current:
                continue

            # Delete surface fields.

            if deletions[key] == "DELETE":
                del current[key]

            # Delete potential fields in deeper dicts.

            elif type(deletions[key]) == dict:
                self._delete(deletions[key], current[key])

            # If we've specified a list for and are operating on a list,
            # then eliminate those.

            elif type(deletions[key]) == list and\
                    type(current[key]) == list:

                log.debug("Deleting items from list lists:")
                log.debug("\\%s", deletions[key])
                log.debug("\\%s", current[key])

                for item in deletions[key]:
                    if item in current[key]:
                        current[key].remove(item)

    def delete(self, deletions):
        self._delete(deletions, self.json)

    def _merge(self, change, current):
        for key in list(change.keys()):
            # Move over missing keys

            if key not in current:
                current[key] = change[key]

            # Merge subsequent dicts, or overwrite if wrong types
            # (in a well-behaved merge that shouldn't happen)

            elif type(change[key]) == dict:
                if type(current[key]) != dict:
                    log.warn("Type change! Old value of ['%s'] not dict." %
                            (key,))
                    current[key] = change[key]
                else:
                    self._merge(change[key], current[key])

            # Merge lists (append change items not present in current and
            # potentially change their order based on the contents in change).

            elif type(change[key]) == list:
                if type(current[key]) != list:
                    log.warn("Type change! Old value of ['%s'] not list." %
                            (key, ))
                    current[key] = change[key]
                else:
                    log.debug("Merging lists:")
                    log.debug("\\%s", change[key])
                    log.debug("\\%s", current[key])

                    # Any items not in change are prepended.  This allows the
                    # simple n-item append to work as expected, it allows the
                    # sort case to work as expected, and gives consistent
                    # behavior in the case of items unaccounted for in change.

                    current[key] = [ i for i in current[key] if i not in change[key] ] +\
                            change[key]

            # Move over present

            else:
                current[key] = change[key]

    def merge(self, newconfigs):
        self._merge(newconfigs, self.json)

    def write(self):
        try:
            f = codecs.open(self.filename, "wb", locale_enc)
            json.dump(self.json, f, ensure_ascii=False, sort_keys=True, indent=4)
        finally:
            f.close()

config = CantoConfig()



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

from .protocol import CantoSocket
from .hooks import call_hook

import logging
import fcntl
import errno
import time
import sys
import os

log = logging.getLogger("CLIENT")

class CantoClient(CantoSocket):
    def __init__(self, socket_name, **kwargs):
        kwargs["server"] = False
        CantoSocket.__init__(self, socket_name, **kwargs)

    def connect(self):
        conn = CantoSocket.connect(self)
        call_hook("client_new_socket", [conn])
        return conn


    # Test whether we can lock the pidfile, and if we can, fork the daemon
    # with the proper arguments.

    def start_daemon(self):
        pidfile = self.conf_dir + "/pid"
        if os.path.exists(pidfile) and os.path.isfile(pidfile):
            try:
                pf = open(pidfile, "a+")
                fcntl.flock(pf.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                if os.path.exists(self.socket_path):
                    os.unlink(self.socket_path)
                fcntl.flock(pf.fileno(), fcntl.LOCK_UN)
                pf.close()
            except IOError as e:
                if e.errno == errno.EAGAIN:
                    # If we failed to get a lock, then the daemon is running
                    # and we're done.
                    return

        pid = os.fork()
        if not pid:
            # Shutup any log output before canto-daemon sets up it's log
            # (particularly the error that one is already running)

            fd = os.open("/dev/null", os.O_RDWR)
            os.dup2(fd, sys.stderr.fileno())

            cmd = "canto-daemon -D " + self.conf_dir
            if self.verbosity > 0:
                cmd += " -" + ("v" * self.verbosity)

            os.setpgid(os.getpid(), os.getpid())
            os.execve("/bin/sh", ["/bin/sh", "-c", cmd], os.environ)

            # Should never get here, but just in case.
            sys.exit(-1)

        while not os.path.exists(self.socket_path):
            time.sleep(0.1)

        return pid

    # Write a (cmd, args)
    def write(self, cmd, args, conn=0):
        return self.do_write(self.sockets[conn], cmd, args)

    # Read a (cmd, args)
    def read(self, timeout=None, conn=0):
        return self.do_read(self.sockets[conn], timeout)



# -*- coding: utf-8 -*-

#Canto - RSS reader backend
#   Copyright (C) 2016 Jack Miller <jack@codezen.org>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License version 2 as 
#   published by the Free Software Foundation.

import logging

log = logging.getLogger("FORMAT")

def get_formatter(fmt, keys):
    l = len(fmt)
    def formatter(dct):
        s = ""
        i = 0
        while i < l:
            if fmt[i] == '%':
                i += 1
                code = fmt[i]
                if code not in keys:
                    i += 1
                    continue
                key = keys[code]
                if key not in dct:
                    i += 1
                    continue
                s += repr(dct[key])
            elif fmt[i] == '\\':
                s += fmt[i + 1]
                i += 1
            else:
                s += fmt[i]
            i += 1
        return s
    return formatter

def escsplit(arg, delim, maxsplit=0, minsplit=0, escapeterms=False):
    r = []
    acc = ""
    escaped = False
    skipchars = 0

    for i, c in enumerate(arg):
        if skipchars > 0:
            skipchars -= 1
            continue

        if escaped:
            escaped = False
            acc += c

        # We append the escape character because we just want to intelligently
        # split, not unescape the components

        elif c == '\\':
            escaped = True

            # Don't unescape things that may need to be split again. Most
            # notably canto-remote splitting on = and then on .

            if not escapeterms:
                acc += c

        elif c == delim[0]:

            # If this is a multi-char delimiter that doesn't match
            # keep the character and move on.

            if len(delim) > 1 and\
                    arg[i : i+len(delim)] != delim:
                acc += c
                continue

            # If we have matched a 1 or multi-char delimiter we need
            # to skip the remaining characters.

            else:
                skipchars = len(delim) - 1

            r.append(acc)
            acc = ""

            # Last split?
            if maxsplit == 1:
                r.append(arg[i + 1:])
                break
            elif maxsplit > 1:
                maxsplit -= 1
        else:
            acc += c
    else:
        # Get last frag, if we didn't maxout.
        r.append(acc)

    if minsplit > 0 and len(r) < (minsplit + 1):
        r += [ None ] * ((minsplit + 1) - len(r))

    return r



# Canto Reddit Plugin
# by Jack Miller
# v1.2
#
# If this is placed in the plugins directory, it will add a new sort:
# reddit_score_sort, and will add "score [subreddit]" to the beginning of
# every relevant feed item.

# PREPEND_SCORE, if true will add the score to the entry title. Note, this
# doesn't effect the sort.

PREPEND_SCORE = True

# PREPEND_SUBREDDIT, if true will add the [subreddit] to the entry title.

PREPEND_SUBREDDIT = True

# EXTRA_LOG_OUTPUT, if true will log non-error stuff with -v.

EXTRA_LOG_OUTPUT = False

# You shouldn't have to change anything beyond this line.

from canto_next.plugins import check_program

check_program("canto-daemon")

from canto_next.fetch import DaemonFetchThreadPlugin
from canto_next.feed import DaemonFeedPlugin
from canto_next.transform import transform_locals, CantoTransform 

import urllib.request, urllib.error, urllib.parse
import logging
import time
import json
import re

log = logging.getLogger("REDDIT")

def debug(message):
    if EXTRA_LOG_OUTPUT:
        log.debug(message)

keep_attrs = [ "score", "subreddit" ]

class RedditFetchJSON(DaemonFetchThreadPlugin):
    def __init__(self, fetch_thread):
        self.plugin_attrs = {
                "fetch_redditJSON" : self.fetch_redditJSON,
        }

        self.comment_id_regex = re.compile(".*comments/([^/]*)/.*")
        self.tb_id_regex = re.compile(".*tb/([^/]*)")

    def fetch_redditJSON(self, **kwargs):
        if "reddit.com" not in kwargs["feed"].URL:
            return

        # Get the feed's JSON
        try:
            json_url = kwargs["feed"].URL.replace("/.rss","/.json")
            req = urllib.request.Request(json_url, headers = { "User-Agent" : "Canto-Reddit-Plugin"})
            response = urllib.request.urlopen(req, None, 10)
            reddit_json = json.loads(response.read().decode())
        except Exception as e:
            log.error("Error fetching Reddit JSON: %s" % e)
            return

        for entry in kwargs["newcontent"]["entries"]:
            m = self.comment_id_regex.match(entry["link"])
            if not m:
                m = self.tb_id_regex.match(entry["link"])
            if not m:
                debug("Couldn't find ID in %s ?!" % entry["link"])
                continue
            m = "t3_" + m.groups()[0]

            for rj in reddit_json["data"]["children"]:
                if rj["data"]["name"] == m:
                    debug("Found m=%s" % m)

                    d = { "data" : {}}
                    for attr in keep_attrs:
                        if attr in rj["data"]:
                            d["data"][attr] = rj["data"][attr]

                    entry["reddit-json"] = d
                    break
            else:
                debug("Couldn't find m= %s" % m)

class RedditScoreSort(CantoTransform):
    def __init__(self):
        self.name = "Reddit Score Sort"

    def needed_attributes(self, tag):
        return [ "reddit-score" ]

    def transform(self, items, attrs):
        scored = []
        unscored = []

        for item in items:
            if "reddit-score" in attrs[item]:

                # For some reason, reddit-score has been parsed as a string
                # some times. Attempt to coerce.

                if not type(attrs[item]["reddit-score"]) == int:
                    try:
                        attrs[item]["reddit-score"] =\
                                int(attrs[item]["reddit-score"])
                    except:
                        unscored.append(item)
                    else:
                        scored.append(item)
                else:
                    scored.append(item)
            else:
                unscored.append(item)

        scored = [ (attrs[x]["reddit-score"], x) for x in scored ]
        scored.sort()
        scored.reverse()
        scored = [ x for (s, x) in scored ]

        return scored + unscored

class RedditAnnotate(DaemonFeedPlugin):
    def __init__(self, daemon_feed):

        self.plugin_attrs = {
                "edit_reddit" : self.edit_reddit,
        }

    def edit_reddit(self, feed, newcontent, tags_to_add, tags_to_remove, remove_items):
        for entry in newcontent["entries"]:
            if "reddit-json" not in entry:
                debug("NO JSON, bailing")
                continue

            rj = entry["reddit-json"]
            if not rj:
                debug("JSON empty, bailing")
                continue

            if "subreddit" not in entry:
                entry["subreddit"] = rj["data"]["subreddit"]
                if PREPEND_SUBREDDIT:
                    entry["title"] =\
                            "[" + entry["subreddit"] + "] " + entry["title"]

            if PREPEND_SCORE:
                score = rj["data"]["score"]
                if "reddit-score" in entry:
                    entry["title"] = re.sub("^\d+ ", "", entry["title"])

                entry["reddit-score"] = score
                entry["title"] =\
                        ("%d " % entry["reddit-score"]) + entry["title"]

        return (tags_to_add, tags_to_remove, remove_items)

transform_locals["reddit_score_sort"] = RedditScoreSort()
